""" PhenoAI allows users to import and use trained sklearn estimators and keras models out-of-the-box in existing workflows. Once loaded, PhenoAI implements methods to call and compare multiple AInalyses and run PhenoAI in a server-client structure.

Generalization over machine learning methods is a key concept of PhenoAI: it depends as little as possible on the loaded estimator. Although PhenoAI as a package is focussed on applications within data analysis in high energy physics, the interface is also usable outside of this field.

To load estimators, they have to be wrapped in a PhenoAI-specific format: an AInalysis. The .maker module allows the user to store estimators in this format."""

# Package meta data
__version__ = "0.1.3"
__versionnumber__ = 4
__versiondate__ = "2018-08-06"
__author__ = "Bob Stienen"
__apiurl__ = "http://www.hef.ru.nl/~bstienen/phenoai/api/update_{}.php"



""" The classes module implements classes from which other classes in the PhenoAI package are derived. Currently these classes are Estimator and Configuration. """

from . import splash
splash.splash()

from copy import deepcopy
import os.path

import numpy as np

from . import exceptions
from . import io
from . import logger
from . import utils



class Estimator(object):
	""" The basic interface methods for estimator classes in the estimators module.
	
	Properties
	----------
	est: estimator
		The estimator to which Estimator derived classes provide an interface. Type of this variable is determined by the derived class.
	path: string
		Path to the stored estimator.
	
	Constructor arguments
	----------------------
	path: string
		Path to the location of the stored estimator.
	load: boolean (default=False)
		Defines if the estimator has to be loaded on initialisation of the interface class.""" 
	def __init__(self, path=None, load=False):
		if not utils.is_none(path):
			if not os.path.isdir(path):
				raise FileNotFoundError("Estimator could not be found at path '{}'.".format(path))
			self.est = None
			if path[-1] == "/":
				path = path[:-1]
			self.path = path
			if load:
				self.load()
	def clear(self):
		""" Removes the loaded estimator from memory.

		This saves memory, but does mean that the estimator has to be reloaded as soon as a new prediction with it has to be made."""
		self.est = None
	def is_loaded(self):
		""" Checks if the estimator is loaded.

		Returns
		-------
		is_loaded: bool
			True if estimator is currently loaded, False otherwise."""
		return (not utils.is_none(self.est))

class Configuration(object):
	""" The basic interface to configuration classes.
	
	This base class defines all loading, resetting and printing functions for configuration classes throughout the package (currently only ainalyses.AInalysisConfiguration). Handles only configuration files following the .yaml format, see http://www.yaml.org.
	
	Properties
	----------
	configuration: dictionary (default={})
		Content of the configuration. Contains an empty dictionary if configuration is not loaded.
	folder: string (default=`None`)
		AInalysis folder. Is deduced from provided path of the AInalysis configuration.
	path: string (default=`None`)
		Location where the .yaml configuration is stored.
	validated: boolean (default=False)
		Boolean indicating if the configuration stored in the configuration property has been validated by the validate method in the derived class.
	
	Constructor parameters
	----------------------
	path: string (default=`None`)
		Location where the .yaml configuration file is stored."""
	def __init__(self, path=None, entries=None):
		self.configuration = {}
		self.path = None
		self.folder = None
		if not utils.is_none(path):
			self.load(path)
		elif not utils.is_none(entries) and isinstance(entries, dict):
			self.configuration = entries
		self.validated = False
	def __getitem__(self, key):
		return self.configuration[key]
	def __setitem__(self, key, value):
		self.configuration[key] = value
	def get(self, key = None):
		""" Returns the configuration in dictionary form or an entry of the configuration

		Parameters
		----------
		key: string (default=`None`)
			Key of the dictionary entry that has to be returned. If no key or `None` was provided the entire dictionary is returned.

		Returns
		-------
		configvalue: dictionary or dictionary entry
			If key was provided as input argument the entry of the configuration with that key is returned. If no key or `None` was provided, the entire configuration directionary is returned."""
		if utils.is_none(key):
			return self.configuration
		else:
			return self.configuration[key]
	def show(self, printdict = None):
		""" Prints the entire configuration via the print() command.

		Parameters
		----------
		printdict: dictionary (default=`None`)
			The dictionary that has to be printed by the print() command. If no dictionary or `None` is provided the self.configuration property will be used instead. """
		if printdict == None:
			printdict = self.configuration
		maxkeylength = 0
		for k in printdict.keys():
			if not isinstance(k, str):
				printdict[str(k)] = printdict[k]
				del(printdict[k])
				k = str(k)
			if len(k) > maxkeylength:
				maxkeylength = len(k)
		for key, value in sorted(printdict.items()):
			if isinstance(value, dict):
				self.show(printdict[key])
			elif isinstance(value, list):
				if len(value) == 1:
					print(key.ljust(maxkeylength+6)+"list of length {}: [{}]".format(len(value), value[0]))
				elif len(value) <= 5:
					print(value)
					liststr = utils.multi_join(value)
					print(key.ljust(maxkeylength+6)+"list of length {}: [{}]".format(len(value), liststr))
				else:
					print(key.ljust(maxkeylength+6)+"list of length {}: [{}, {}, ... {}, {}]".format(len(value), value[0], value[1], value[-2], value[-1]))
			else:
				print(key.ljust(maxkeylength+6)+str(value))
	def load(self, path):
		""" Loads the configuration indicated by the provided path.

		The loaded configuration is stored as a dictionary in the configuration property of the object. The provided path is stored in the path property. Configuration files have to follow the .yaml format (see http://www.yaml.org/)

		Parameters
		----------
		path: string
			Location of the configuration .yaml file. """
		self.configuration = io.read_yaml(path)
		self.path = path
		pathparts = path.split('/')
		folderparts = pathparts[:-1]
		self.folder = "/".join(folderparts)
	def validate(self):
		""" This method is only implemented in derived classes. If called on a Configuration instance a ConfigurationException is raised. """
		raise ConfigurationException("Cannot call validate method on Configuration instance. Only available in instances derived from the Configuration class.")

class AInalysisResults(object):
	""" Container for and interface to results of an AInalysis instance
	
	The AInalysisResults class functions as a container for results produced by an AInalysis instance and provides methods to access these results efficiently.
	
	Properties
	----------
	result_id: string
		ID of the result object. Equal to the AInalysisID of the AInalysis that produced the results stored in this instance.
	configuration: AInalysisConfiguration instance
		Configuration of the AInalysis that produced the results stored in this instance.
	data: numpy.ndarray
		Data that has been subjected to prediction (after possible mapping).
	data_ids: list of strings
		List if unique IDs (strings) by which data points can be identified. If set, data points and their classifications and predictions can be read from the AInalysisResults object by referencing their ID. If not set (i.e. ids is set to `None`) this functionality is not available.
	mapped: list or boolean
		False if no mapping has taken place. If mapping has taken place, this property stores a list of booleans indicating on a per datapoint basis if that specific datapoint has been mapped or not. In that case the data stored in the data property of this AInalysisResults instance is this mapped data.
	predictions: numpy.ndarray
		Results of the prediction method of the estimator stored in the AInalysis object with the data stored in the data property of this AInalysisResults instance.
	
	Constructor parameters
	----------------------
	result_id: string
		ID of the result object. Equal to the AInalysisID of the AInalysis that produced the results stored in this instance.
	configuration: AInalysisConfiguration instance
		Configuration of the AInalysis that produced the results stored in this instance.
	data: numpy.ndarray
		Data that has been subjected to prediction (after possible mapping).
	data_ids: list of strings (default=`None`)
		List if unique IDs (strings) by which data points can be identified. If set, data points and their classifications and predictions can be read from the AInalysisResults object by referencing their ID. If not set (i.e. ids is set to `None`) this functionality is not available.
	mapped: boolean (default=False)
		Boolean indicating if the data has been mapped before prediction. If True, the data stored in the data property of this AInalysisResults instance is this mapped data.
	predictions: numpy.ndarray (default=`None`)
		Results of the prediction method of the estimator stored in the AInalysis object with the data stored in the data property of this AInalysisResults instance. """
	def __init__(self, result_id, configuration, data, data_ids=None, mapped=False, predictions=None):
		self.result_id = result_id
		self.data = data
		self.data_ids = data_ids
		self.mapped = mapped
		self.configuration = Configuration(entries=configuration.configuration)
		self.predictions = predictions
	def get(self, array, reference=None):
		""" Returns the content of the array at location of the reference

		Selects and returns the rows from the provided array which are referenced by the reference argument of this function.

		Parameters
		----------
		array: numpy.ndarray
			Numpy array from which rows have to be selected
		reference: numpy.ndarray or list of references (default=`None`)
			If this reference is an integer, that row of the array with that integer as rownumber will be returned. If the reference is a string, the rownumber of the entry in the self.data_ids equal to this reference will be used instead. The reference variable can also be a list or numpy.ndarray of integers and/or strings. In that case a numpy.ndarray with the selected entries will be returned.

		Returns
		-------
		selection: numpy.ndarray
			Selected row(s) from array corresponding to the provided reference(s). """

		if utils.is_none(reference):
			return array
		elif isinstance(reference, int):
			return array[reference]
		elif isinstance(reference, str):
			for i, did in enumerate(self.data_ids):
				if reference == did:
					return array[i]
		elif isinstance(reference, np.ndarray) or isinstance(reference, list):
			if len(array.shape) == 1:
				a = len(array)
			else:
				a = len(array[0])
			selection = np.zeros(( len(reference), a ))
			for i in range(len(reference)):
				selection[i,:] = self.get(array, reference[i])
			return selection
		raise exceptions.ResultsException("Unknown reference format for get: {}.".format(type(reference)))
	def get_ids(self):
		""" Returns data ids

		Returns
		-------
		ids: numpy.ndarray
			All IDs for the data in this AInalysisResults object"""
		return self.data_ids
	def get_data(self, i=None):
		""" Returns data (rows) used for prediction

		Parameters
		----------
		i: integer, string or list of integers/strings (default=`None`)
			Reference used for selecting rows in the data array. See the documentation for the get method for allowed values and consequences.

		Returns
		-------
		data: numpy.ndarray
			Selected rows from the data array."""
		return self.get(self.data, i)
	def get_classifications(self, i=None, variable=None, calibrated=True):
		""" Returns the classification results based on the predictions stored in the predictions property.

		Uses the information stored in the configuration property to classify the results in the predictions property. These classifications are then returned. If the results are not interpretable as classifications (e.g. because the AInalysis implemented a regressor instead of a classifier) `None` is returned.

		Classification names or indicators or extracted from the 'output' entry in the configuration.

		Parameters
		----------
		i: integer or numpy.ndarray (default=`None`)
			Reference used for generating classifications from the predictions array. See the documentation for the get method for allowed values and consequences.
		calibrated: boolean (default=True)
			Boolean indicating if prediction results have to be calibrated. See documentation for the get_predictions method for requirements.

		Returns
		-------
		classifications: numpy.ndarray
			Requested classifications. """
		if not self.configuration["type"] == "classifier":
			return None
		# Select data
		predictions = self.get(self.predictions, i)
		# Get maxima
		if len(predictions.shape) == 1:
			maxs = np.argmax(predictions, axis=0)
		else:
			maxs = np.argmax(predictions, axis=1)
		# Return classifications
		return self.configuration["classes"][maxs]
	def get_predictions(self, i=None, variable=None, calibrated=True):
		""" Returns the predictions stored in predictions property.

		Returns the prediction results stored in the predictions property of this object. By setting the `calibrated` argument to True (default), the prediction results will first be calibrated before returning, if all of the following requirements are met:

		- the estimator is a classifier;
		- the estimator is not already calibrated;
		- the configuration validly defines a method to calibrate the results.

		If any of these requirements is violated, the bare prediction results are returned. If `calibrated` is set to False, the raw prediction results will be returned either way.

		Parameters
		----------
		i: integer or numpy.ndarray (default=`None`)
			Reference used for selecting rows in the predictions array. See the documentation for the get method for allowed values and consequences.
		calibrated: boolean (default=True)
			Boolean indicating if prediction results have to be calibrated. See documentation for this method for requirements.

		Returns
		-------
		predictions: numpy.ndarray
			Requested (calibrated) predictions. """

		# Get predictions
		preds = self.get(self.predictions, i)
		# Check requirements for calibration
		if calibrated:
			if self.configuration["type"] == 'classifier':
				if not self.configuration["classifier.calibrated"]:
					if self.configuration["classifier.calibrate"]:
						# Get calibration configuration
						bins = self.configuration["classifier.calibrate.bins"]
						values = self.configuration["classifier.calibrate.bins"]
						# Perform calibration
						cut = bins[np.argmin(np.abs(values-0.5))]
						classifications = 1.0*(predictions > cut)
						predscal = np.zeros(len(preds))
						for i in range(len(predictions)):
							b = np.argmin(np.abs(bins - preds[i]))
							predscal[i] = values[b]
						return predscal
		return preds
	def is_outlier(self, i=None, use_map_target_area=False):
		""" Returns information about whether or not data used in prediction lies outside of region sampled with training data.

		Checks for all data selected via reference variable i if it is outside of the region from which training data was sampled for the training of the estimator of the AInalysis that created this AInalysisResults object. By setting the use_map_target_area argument to True the reference area will be taken as the target area for the mapping procedure (if any is defined via a floating point number in the AInalysis configuration). Returns a numpy.ndarray containing booleans indicating if the data points are outside of the sample region (True) or not (False). 

		Note that if mapping took place before prediction (checkable via the is_mapped method) all values returned will be False.

		Parameters
		----------
		i: integer or numpy.ndarray (default=`None`)
			Reference used for selecting data for which the outsiderness has to be determined. See the documentation for the get method for allowed values and consequences.
		use_map_target_area: boolean (default=False)
			Use the mapping target area as reference area instead of the area from which training data was sampled. If set to True but the mapping was not configured via a floating point value in the AInalysis configuration, the method will continue as if this value was set to False.

		Returns
		-------
		outsider: numpy.ndarray
			Numpy array containing booleans. True if data point is an outsider, False otherwise."""

		if not isinstance(self.configuration["mapping"], float) and use_map_target_area:
			raise exceptions.ResultsException("No mapping target area defined in AInalysis configuration.")
		selection_mins = self.get(self.data, i)
		selection_maxs = deepcopy(selection_mins)
		for i in range(len(self.configuration["parameters"])):
			i_min = self.configuration["parameters"][i][2]
			i_max = self.configuration["parameters"][i][3]
			if use_map_target_area:
				r = self.configuration["parameters"][i][3] - self.configuration["parameters"][i][2]
				i_min += r*self.configuration["mapping"]
				i_max -= r*self.configuration["mapping"]
			selection_mins[i] -= i_min
			selection_maxs[i] -= i_max
		outofrange = selection_mins[selection_mins<0]*1.0 + selection_maxs[selection_maxs>0]*1.0
		return np.sum(outofrange) == 1.0
	def is_mapped(self, i=None):
		""" Returns information about whether or not provided data was mapped before prediction took place.

		Parameters
		----------
		i: integer or numpy.ndarray (default=`None`)
			Reference used for selecting data of which mapping information is requested. See the documentation for the get method for allowed values and consequences.

		Returns
		-------
		mapped_global: boolean
			Boolean indicating if mapping procedure has been called at all. Can indicate that map_data was set to False by the user (or left at this default value) or that the AInalysis does not support data mapping.
		mapped: numpy.ndarray of booleans or `None`
			If mapped_global is True this variable will contain a numpy array filled with booleans indicating per data point indicated via reference variable i if it was mapped. This variable will contain `None` if mapped_global was False. """
		if (isinstance(self.mapped, bool) and self.mapped == False) or utils.is_none(self.mapped):
			return (False, None)
		else:
			selection = self.get(self.mapped, i)
			return (True, selection)
	def num(self):
		""" Returns the length of the data (i.e. the number of data points) stored in the AInalysisResults instance.

		Returns
		-------
		L: integer
			Number of data points (and hence the number of predictions) stored in the AInalysisResults instance. """
		return len(self.data)
	def __len__(self):
		return self.num()
	def summary(self):
		""" Prints a summary of the contents of this object """
		logger.debug("Print report for AInalysisResults")
		print("---------------------------")
		print("| AInalysisResults report |")
		print("---------------------------")
		print("ID: {}".format(self.result_id))
		print("Length: {}".format(self.num()))
		# Information on data
		contains_data = not isinstance(self.get_data(), type(None))
		print("Contains data: {}".format(contains_data))
		if contains_data:
			print("   Data shape: {}".format(self.get_data().shape))
			print("   Access via .get_data( args )")
		# Information on Data ids
		contains_data_ids = not isinstance(self.get_ids(), type(None))
		print("Contains data IDs: {}".format(contains_data_ids))
		if contains_data_ids:
			print("   Data IDs length: {}".format(len(self.get_ids())))
			print("   Get full list via .get_ids()")
			print("   Use as argument in .get_*() methods")
		# Information on data mapping
		data_is_mapped = self.is_mapped()[0]
		print("Data is mapped: {}".format(data_is_mapped))
		if data_is_mapped:
			if isinstance(self.is_mapped()[1], list):
				print("   Data IDs length: {}".format(len(self.is_mapped()[1])))
			elif isinstance(self.is_mapped()[1], np.ndarray):
				print("   Data IDs shape: {}".format(self.is_mapped()[1].shape))
			print("   Access via .is_mapped( args )")
		# Information on Data ids
		contains_predictions = not isinstance(self.get_predictions(), type(None))
		print("Contains predictions: {}".format(contains_predictions))
		if contains_predictions:
			print("   Predictions shape: {}".format(self.get_predictions().shape))
			print("   Access via .get_predictions( args )")

class PhenoAIResults(object):
	""" Container class for AInalysisResults.
	
	This class functions as a container class for AInalysisResults from multiple AInalyses. Methods are implemented to aid accessing these instances.
	
	Properties
	----------
	results: list
		List of AInalysisResults. """
	def __init__(self):
		self.results = []
	def add(self, result):
		""" Appends an AInalysisResults instance to the container.

		An error is raised if the result that is to be appended has an AInalysisID that is already present in the stored AInalysisResults instances.

		Parameters
		----------
		result: AInalysisResults instance
			AInalysisResults instance that has to be added to the container. """
		if not utils.is_none(self.get(result.result_id)):
			raise exceptions.ResultsException("Already an AInalysisResults instance stored with AInalysisID '{}'".format(result.result_id))
		self.results.append(result)
	def num(self):
		""" Returns the number of stored AInalysisResults in this container.

		Returns
		-------
		n: integer
			Number of AInalysisResults stored in this container. """
		return len(self.results)
	def __len__(self):
		return self.num()
	def get(self, result_id):
		""" Returns the AInalysis with a given AInalysisID.

		Looks in stored results if there is an AInalysisResults instance with provided ID as Result ID. If found, this instance is returned. If none is found, the provided id is checked if it is an integer in range 0 ... len(results)-1. If so, the result with that internal index is returned (corresponding to the ID from get_ids() at that index). If no such result could be returned, `None` is returned instead.

		Parameters
		----------
		result_id: string
			The Result ID of the AInalysisResults that has to be returned.

		Returns
		-------
		result: AInalysisResults instance / `None`
			AInalysisResults instance with the requested AInalysisID. If no such instance was found, `None` is returned. """
		for i in range(len(self.results)):
			if self.results[i].result_id == result_id:
				return self.results[i]
		if isinstance(result_id, int) and result_id >= 0 and result_id < len(self.results):
			return self.results[result_id]
		return None
	def __getitem__(self, result_id):
		return self.get(result_id)
	def get_ids(self):
		""" Returns a list of the ResultIDs of all stored AInalysisResults instances.

		Returns
		-------
		ids: list
			List of all ResultIDs fo all stored AInalysisResults instances. """
		ids = []
		for i in range(len(self.results)):
			ids.append(self.results[i].result_id)
		return ids
	def summary(self):
		""" Prints a summary of the contents of this object """
		logger.debug("Print report for AInalysisResults")
		print("===========================")
		print("||  PhenoAIResults report  ||")
		print("===========================")
		print("Contains {} AInalysisResults object(s)".format(self.num()))
		print("AInalysisResult IDs:")
		for i in range(self.num()):
			print("  - {} (length: {})".format(self.get_ids()[i], self.get(self.get_ids()[i]).num()))
		print("Access AInalysisResults objects by id via PhenoAIResults[ id ]")


""" The io module contains functions to read and create files. Files that have a reader interface in this module are .slha, .hdf5, .csv, .sfv and .yaml. Files with a writer interface are .hdf5, .csv and .yaml. A more global function to find all files with a specific extension is implemented via the get_file_paths function. """ 
from os import listdir
import os.path
import codecs
try:
	import cPickle as pkl
except:
	import pickle as pkl

import numpy as np
import pyslha
from ast import literal_eval
import h5py
from yaml import load as yamlload
try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from . import exceptions
from . import logger
from . import utils



def get_file_paths(locations, extensions=None, recursive=False):
	""" Return all paths from all files in a folder fulfilling set requirements

	This function walks over all files in a folder. If its extension matches the extension requirement (if any) the path to the file is stored in a list which is returned by this function. By using the `recursive` argument to this function all encountered folders will be walked through as well.

	Parameters
	----------
	locations: string or list of strings
		Paths to folders which have to be walked through looking for files
	extensions: string or list of strings (default=`None`)
		Extensions of files of which the path has to be returned. If `None` all encountered files will have their path stored.
	recursive: boolean
		Determines if encountered folders have to be walked through as well.

	Returns
	-------
	locations: list of strings
		List of all paths to files fulfilling the extensions requirement."""


	# Locations of the found files
	locs = []
	# If provided locations argument is a list call this function for each of the entries
	# Add result of each call to list of found file paths
	if isinstance(locations, list):
		for l in locations:
			locs.extend(get_file_paths(l, extensions))
	# Check if provided locations argument is a string
	elif isinstance(locations, str):
		# Check if locations is a valid location
		if not os.path.exists(locations):
			raise exceptions.FileIOException("Location '{}' does not exist.".format(locations))
		# Check if it is a directory
		if os.path.isdir(locations):
			# If search is resursive get files in subdir and call this function recursively
			# Add result for each call to list of found file paths
			content = [os.path.join(locations, f) for f in listdir(locations) if os.path.isfile(os.path.join(locations, f))]
			for c in content:
				if os.path.isdir(c) and recursive:
					paths = get_file_paths(c, extensions, True)
					if paths != None:
						locs.extend(paths)
				elif os.path.isfile(c):
					path = get_file_paths(c, extensions, recursive)
					if path != None:
						locs.extend(path)
		elif os.path.isfile(locations):
			if extensions == None:
				# No requirement, add location to list of file paths
				locs = [locations]
			elif isinstance(extensions, str):
				# Requirement is a string, check if file name ends with extension
				# If it does, add file path to list of found file paths
				if locations.endswith(extensions):
					locs.append(locations)
			elif isinstance(extensions, list):
				# Requirement is a list, check for each of the entries in the list if it contains the extension of the file
				# If it does, add file path to list of found file paths
				for ext in extensions:
					if locations.endswith(ext):
						locs.append(locations)
						break
			else:
				# No valid extension requirement provided
				raise exceptions.FileIOException("Extension of files has to be a (list of) string(s), not a '{}'.".format(type(extensions)))
		else:
			return
	else:
		# No valid locations argument input
		raise exceptions.FileIOException("Location of files has to be a (list of) path(s) [string] to files or folders containing them. Provided was '{}'.".format(type(locations)))
	# Return list of file locations
	locs = sorted(locs)
	return locs

def read_slha(path, reader_list=None):
	""" Reads a .slha file to pyslha.Doc object or numpy.ndarray

	Reads the content of a .slha file into a pyslha.Doc. If a reader list was provided these requested entries are extracted fromt the pyslha.Doc object and returned as a numpy.ndarray. If no readerlist was provided the pyslha.Doc object is returned.

	Parameters
	----------
	path: string
		Path of the .slha file to be read
	reader_list: list of slha [BLOCK, SWITCH] entries
		List of entries in the .slha file, denoted by [BLOCK, SWITCH] entries. A reader list is therefore 2-dimensional object (a list of lists).

	Returns
	-------
	content: pyslha.Doc or numpy.ndarray
		If a valid reader list was provided a numpy.ndarray with the requested content is returned. Else the pylsha.Doc object is returned. """

	# Read .slha file into pyslha.Doc object
	docobj = None
	if(isinstance(path, pyslha.Doc)):
		return path
	elif isinstance(path, str):
		if os.path.isfile(path):
			try:
				docobj = pyslha.read(path)
			except Exception as e:
				raise exceptions.FileIOException("Unexpected pyslha error while reading '{}': {}.".format(path, str(e)))
		else:
			raise exceptions.FileIOException("File not found '{}'.".format(path))
	else:
		raise exceptions.FileIOException("Filepath to the .slha to read has to be a string (supplied: '{}').".format(type(path)))

	# Check if reader list is provided
	if utils.is_none(reader_list):
		return docobj
	elif isinstance(reader_list, list):
		data = np.zeros(len(reader_list))
		for i in range(len(reader_list)):
			if len(reader_list[i]) != 2:
				raise exceptions.FileIOException("Datalist must only contain lists with format [BLOCK, SWITCH].")
			try:
				try:
					data[i] = docobj.blocks[reader_list[i][0].upper()][int(reader_list[i][1])]
				except:
					data[i] = docobj.blocks[reader_list[i][0].upper()][reader_list[i][1]]
			except ValueError as e:
				raise exceptions.FileIOException("Switch '{}' could not be casted to an integer or tuple.".format(reader_list[i][1]))
			except KeyError:
				raise exceptions.FileIOException("No Switch '{}' in Block '{}' found in .slhafile.".format(reader_list[i][1], reader_list[i][0]))
		return data
	else:
		return docobj

def read_hdf5(path, name):
	""" Reads hdf5 file to numpy array

	Parameters
	----------
	path: string
		Path to the .hdf5 file.
	name: string
		Name of the array to be read.

	Returns
	-------
	content: numpy.ndarray
		Content of the .hdf5 file. """
	with h5py.File(path, 'r') as f:
		content = f[name][()]
	return content
def read_csv(path, headerlines=False):
	""" Reads and returns the content of a .csv file

	Parameters
	----------
	path: string
		Path to the .csv file to be read.
	headerlines: boolean or integer (default=False)
		If set to an integer, that number of first lines is not included in the content array. If set to True, the first line is excluded. No lines are excluded from the content array if set to False.

	Returns
	-------
	content: numpy.ndarray
		Content of the .csv file. """

	content = genfromtxt(path, delimiter=',')
	if headerlines:
		if isinstance(headerlines, int):
			content = content[:headerlines]
		else:
			content = content[:headerlines]

	return content
def read_checksum(path):
	""" Reads checksums from a checksum file and returns them in a dictionary

	Parameters
	----------
	path: string
		Path to a checksum file (.sfv)

	Returns
	-------
	checksums: dictionary
		Dictionary of checksums stored in the provided checksum file. """
	if not os.path.isfile(path):
		raise exceptions.FileIOException("Checksum file '{}' could not be found.".format(path))
	checksums = {}
	with open(path, 'r') as csfile:
		lines = csfile.readlines()
		for l in lines:
			word = l.split()
			checksums[word[0]] = word[1]
	return checksums
def read_yaml(path):
	""" Reads the content of a .yaml file returns it as dictionary.

	Parameters
	----------
	path: string
		Path to the .yaml file to be read.

	Returns
	-------
	yamlcontent: dictionary
		Content of the read .yaml file in dictionary format. """
	if not os.path.isfile(path):
		raise ImportError("YAML file could not be found at location '{}'.".format(path))
	with open(path) as runcard:
		yamlcontent = yamlload(runcard, Loader=Loader)
	return yamlcontent

def get_lsp_from_slha(slha):
	""" Reads the LSP(s) for a (list of) .slha file(s)

	Finds and returns the switch of the lightest supersymmetric particle in the MASS block of a .slha file. As such a minimal switch value of 1000000 is required. If multiple particles could all be qualified as the LSP, the one with the smallest switch value is given.

	Parameters
	----------
	slha: string or list of strings
		Path to a .slha file or a list of paths to .slha files.

	Returns
	-------
	lsps: integer or list of integers
		The switch of the LSP in the provided .slha file. If a list of .slha files was provided a list of switches is returned. """

	if isinstance(slha, list):
		correct_lsp = [None]*len(slha)
		for i, s in enumerate(slha):
			correct_lsp[i] = read_lsp(s)
		return correct_lsp
	elif isinstance(slha, str) or isinstance(slha, pyslha.Doc):
		spectrum = load_slha(slha)
		blocks = [spectrum.blocks[i].name for i in spectrum.blocks]
		if 'MASS' in blocks:
			masses = np.array([[i, spectrum.blocks['MASS'][i]] for i in spectrum.blocks['MASS'].keys() if i>1e6])
			masses = masses[abs(masses[:,1]).argsort()]
			return masses[0,0]
		else:
			raise exceptions.FileIOException("No MASS block found, so no LSP check could be performed.")
	else:
		raise exceptions.FileIOException("Can only read the LSP of a .slha file (provide string of path to file) or on a list of .slha files. Provided was a '{}'".format(type(slha)))

def write_hdf5(path, name, nparray):
	""" Writes a numpy array to a file in .hdf5 format.

	Parameters
	----------
	path: string
		Location to where the .hdf5 should be written.
	name: string
		Name of the array in the .hdf5 file
	nparray: numpy.ndarray
		Numpy array that should be saved in the .hdf5 file. """
	with h5py.File(path, 'w') as hf:
		hf.create_dataset(name,  data=nparray)
def write_csv(path, nparray, header=None):
	""" Writes a numpy array to a file in .csv format.

	Parameters
	----------
	path: string
		Location to where the .csv should be written.
	nparray: numpy.ndarray
		Numpy array that should be saved in the .csv file.
	header: list or string (default=`None`)
		Header line at the top of the .csv file. Can be a string (which will form the literal top line of the .csv file) or a list (which will be joined with "," to form the top line). If `None` no header line is inserted. """

	if isinstance(header, list):
		header = ",".join(header)
	if isinstance(header, str):
		numpy.savetxt(path, nparray, delimiter=",", header=header)
	elif utils.is_none(header):
		numpy.savetxt(path, nparray, delimiter=",")
	else:
		raise FileIOException("Provided header was not a string or a list")
def write_yaml(path, dictionary):
	""" Writes a dictionary to a file in .yaml format.

	Parameters
	----------
	path: string
		Location to where the .yaml should be written.
	dictionary: dictionary
		Dictionary that should be stored in the file."""
	with open(path) as f:
		yaml.dump(dictionary, f)

def pickle(obj):
	""" Creates a serialization of the provided object

	Serialization is done by pickle module. If cPickle package is available, that package will be used instead, yielding a gain in speed.

	Parameters
	----------
	obj: Object
		Object to be serialized.

	Returns
	-------
	pickle: pickle object
		Serialized version of the provided object. """
	return codecs.encode(pkl.dumps(obj), "base64").decode()
def unpickle(pickle):
	""" Creates an object from a serialized one

	Unserialization is done by pickle module. If cPickle package is available, that package will be used instead, yielding a gain in speed.

	Be aware that unpickling a serialized object or file from an unknown source yields safety risks: serialized files are just compressed computer code that can be executed.

	Parameters
	----------
	obj: Object
		Object serialized with pickle (or cPickle) package.

	Returns
	-------
	pickle: pickle object
		Unserialized version of provided object. """
	return pkl.loads(codecs.decode(pickle.encode(), "base64"))



""" Miscellaneous functions

This module implements various functions that are used throughout the package and serve general purposes. """

import random
import string
import os
import zlib
from decimal import Decimal

import numpy as np


def is_none(x):
	""" Checks if variable equals `None`

	This is useful, because comparing a numpy ndarray to `None` yields a warning or an error (depening on installed version of numpy)

	Parameters
	----------
	x: any variable
		Variable that needs to be checked for being equal to `None`

	Returns
	-------
	b: boolean
		True of input variable equals `None`, else False. """
	return isinstance(x, type(None))
def random_string(length):
	""" Creates a random string with specified length

	Returned string contains uppercase characters, lowercase characters and digits

	Parameters
	----------
	length: integer
		Length of the returned string

	Returns
	-------
	rndstr: string
		Random string with specified length. """
	return ''.join(random.choices(string.ascii_uppercase + string.digits + string.ascii_lowercase, k=length))
def bool_to_text(b):
	""" Changes a provided boolean into a 'Yes' or 'No' string

	No test is performed on whether or not the provided variable is boolean or not, only an 'if b: / else:' test is performed.

	Parameters
	----------
	b: any object
		Object that will be tested for its evaluation to a boolean.
		
	Returns
	-------
	text: string
		'Yes' if input variable is evaluated as `True`, 'No' otherwise. """
	if b:
		return "Yes"
	else:
		return "No"
def stringify(l):
	""" Converts all entries in a list or array to strings

	Parameters
	----------
	l: list or numpy.ndarray
		List or numpy ndarray that has to be converted to strings. 

	Returns
	-------
	strings: list or numpy.ndarray
		List or numpy ndarray with strings for entries. """

	# Return stringified version of numpy array
	if isinstance(l, np.ndarray):
		return l.astype(np.str)
	# Raise exception if not a list
	if not isinstance(l, list):
		raise Exception("Can only stringify lists and numpy.ndarrays")
	strings = [None]*len(l)
	for i in range(len(l)):
		# Check if element not a list itself
		if isinstance(l[i], list):
			strings[i] = stringify(l[i])
		else:
			# Stringify element
			strings[i] = str(l[i])
	# Return list of strings
	return strings
def multi_join(l):
	""" Converts a multi-dimensional list into a string

	Parameters
	----------
	l: list or numpy.ndarray
		List that has to be converted into a string.

	Returns
	-------
	s: string
		String of the provided list. """

	content = ""
	for k in range(len(l)):
		if isinstance(l[k], list):
			item = multi_join(l[k])
		else:
			item = str(l[k])
		if content != "":
			content += ", "
		content += item
	return "[{}]".format(content)
def dict_to_matrix(dictionary, nrows, ncols):
	""" Converts a dictionary to a 2-dimensional numpy ndarray 

	Creates a numpy ndarray with specified number of rows and columns. This matrix will be filled with the entries in the dictionary. If the dictionary contains too few entries to fill the matrix, the matrix is zero-padded. An Exception will be raised if the dictionary contains too many entries.

	Parameters
	----------
	dictionary: dictionary
		Dictionary containing the entries with which the matrix has to be filled.
	nrows: integer
		Number of rows for the to-be-generated matrix.
	ncols: integer
		Number of columnns for the to-be-generated matrix.

	Returns
	-------
	matrix: numpy.ndarray
		Numpy ndarray of specified size with entries from provided dictionary. """
	if len(dictionary) > nrows*ncols:
		raise Exception("Dictionary has more entries than the matrix would have with shape (nrows, cols).")
	matrix = np.zeros((int(nrows), int(ncols)))
	for i in range(nrows):
		for j in range(ncols):
			matrix[i,j] = dictionary[i+1,j+1]
	return matrix
def dict_to_square_matrix(dictionary, N=None):
	""" Converts a dictionary to a 2-dimensional square numpy ndarray

	Creates a numpy ndarray with specified number of rows and columns. If no number of rows and columns is provided, this number will be set to the square root of the number of entries in the dictionary (unless this is not an integer, in which case an Exception will be raised).

	This matrix will be filled with the entries in the dictionary. If the dictionary contains too few entries to fill the matrix, the matrix is zero-padded. An Exception will be raised if the dictionary contains too many entries.

	Parameters
	----------
	dictionary: dictionary
		Dictionary containing the entries with which the matrix has to be filled.
	N: integer (default: `None`)
		Number of rows and columns for the square matrix. If `None`, the square root of the number of entries in the dictionary will be used instead.

	Returns
	-------
	matrix: numpy.ndarray
		Numpy ndarray of specified size with entries from provided dictionary. """
	if is_none(N):
		N = np.sqrt(len(dictionary))
		if not isinstance(N, int) or N < 1:
			raise Exception("Dictionary has a non-integer square root.")
	return dict_to_matrix(dictionary, N,N)

# Checksum functions
def convert_to_hex(number, n=9):
	""" Convert number to hexadecimal notation
	
	Parameters
	----------
	number: integer
		Number that has to be converted.
	n: integer (default=9)
		Length of the hexadecimal number.

	Returns
	-------
	hex: string
		Hexadecimal representation of provided number. """
	return format(number & 0xFFFFFFFFF, '0{}x'.format(n))
def calculate_file_checksum(filepath, format_checksum=True):
	""" Calculate the checksum of specified file

	Checksum is calculated with zlib.crc32.

	Parameters
	----------
	filepath: string
		Path to the file for which the checksum has to be calculated
	format_checksum: boolean (default: True)
		Boolean indicating if the checksum has to be formatted to hexadecimal notation

	Returns
	-------
	checksum: integer or string
		Calculated checksum. If format_checksum was True, a hexadecimal string representation of integer checksum is returned. If False, this integer is returned. """
	buffersize = 65536
	crcvalue = 0
	if os.path.exists(filepath):
		with open(filepath, 'rb') as afile:
			buffr = afile.read(buffersize)
			while len(buffr) > 0:
				crcvalue = zlib.crc32(buffr, crcvalue)
				buffr = afile.read(buffersize)
	if format_checksum:
		return convert_to_hex(crcvalue)
	return crcvalue
def calculate_folder_checksum(folderpath, format_checksum=True, recursive=True):
	""" Calculate the checksum of a specified folder

	Checksum is calculated with zlib.crc32.

	Parameters
	----------
	folderpath: string
		Path to the folder for which the checksum has to be calculated
	format_checksum: boolean (default: True)
		Boolean indicating if the checksum has to be formatted to hexadecimal notation
	recursive: boolean (default: True)
		Boolean indicating if the folder has to be searched resursively, also making checksums of files in folders in the folder (depth: unlimited).

	Returns
	-------
	checksum: integer or string
		Calculated checksum. If format_checksum was True, a hexadecimal string representation of integer checksum is returned. If False, this integer is returned. """

	buffersize = 65536
	crcvalue = 0
	if os.path.exists(folderpath):
		for root, subdirs, files in os.walk(folderpath):
			for f in files:
				if f != 'checksums.sfv':
					crcvalue += calculate_file_checksum("{}/{}".format(root,f), False)
			if recursive:
				for s in subdirs:
					crcvalue += calculate_folder_checksum(folderpath+"/"+s, False)
	if format_checksum:
		return convert_to_hex(crcvalue)
	return crcvalue
def calculate_ainalysis_checksums(folder):
	""" Calculate checksums of specified AInalysis

	Calculates all relevant checksums for an AInalysis, namely for

	- the estimator.pkl or estimator.hdf5 file
	- the configuration.yaml file
	- the functions.py file
	- the AInalysis folder as a whole (recursive=True)

	Results are returned as a dictionary of hexadecimal representations of the checksums.

	Parameters
	----------
	folder: string
		Path to the AInalysis folder.

	Returns
	-------
	checksums: dictionary
		Dictionary containing the checksums of the AInalysis folder (see above for specification). """
	if os.path.isfile(folder+"/estimator.pkl"):
		cs_estimator = calculate_file_checksum(folder+"/estimator.pkl")
	else:
		cs_estimator = calculate_file_checksum(folder+"/estimator.hdf5")
	cs_card = calculate_file_checksum(folder+"/configuration.yaml")
	cs_functions = calculate_file_checksum(folder+"/functions.py")
	cs_total = calculate_folder_checksum(folder)
	return {"estimator":cs_estimator, "configuration.yaml":cs_card, "functions.py":cs_functions, "total":cs_total}


""" The estimators module contains interfaces to estimators of different types and a factory class (EstimatorFactory) to create the estimators. The estimators all implement the same sklearn-inspired interface and all inherit from the classes.Estimator class. Currently implemented are the SklearnEstimator and KerasTFEstimator classes (for scikit-learn estimators and keras tensorflow estimators respectively). """

from . import splash
splash.splash()

import os
try:
	import cPickle as pkl
except:
	import pickle as pkl
import threading as t

from . import containers
from . import io
from . import exceptions
from . import logger
from . import utils


class EstimatorFactory(object):
	""" Factory class for implemented estimators. Contains only a single method 'create_estimator', with which new estimators of an indicated type can be created. """
	def create_estimator(self, estimator_type, path):
		""" Creates and returns an estimator of indicated type that can be found at indicated path.

		Parameters
		----------
		estimator_type: string
			Type of the estimator that has to be returned. Can be either "sklearnestimator" or "kerastfestimator" to create a scikit-learn estimator from pickle or keras estimator from hdf5 file respectively.
		path: string
			Path to the estimator.

		Returns
		-------
		estimator: Estimator
			Estimator derived class of indicated estimator type. """
		if estimator_type == "sklearnestimator":
			logger.debug("Create SklearnEstimator in EstimatorFactory instance")
			return SklearnEstimator(path)
		elif estimator_type == "kerastfestimator":
			logger.debug("Create KerasTFEstimator in EstimatorFactory instance")
			return KerasTFEstimator(path)

class SklearnEstimator(containers.Estimator):
	""" Interface to a scikit-learn estimator. Inherits its properties from the containers.Estimator class."""
	def __init__(self, path=None, load=False):
		try:
			from sklearn import __version__ as sklversion
		except:
			logger.critical("Could not create Estimator based on scikit-learn, package sklearn was not installed.")
			raise exceptions.PhenoAIException("Could not create Estimator based on scikit-learn, package sklearn was not installed.")
		self.libraries = {"sklearn": [sklversion]}
		super().__init__(path, load)
	def load(self):
		""" Loads the estimator into the self.est property from the location stored in self.path. """
		logger.debug("Loading estimator")
		with open(self.path+"/estimator.pkl", 'rb') as f:
			self.est = pkl.load(f)
	def save(self, location):
		""" Saves the estimator to provided location.

		Estimator will be stored at specified location as "estimator.pkl"

		Parameters
		----------
		location: string
			Location where the estimator should be stored, including name and extension of the estimator."""
		if location[-1] == "/":
			location = location[:-1]
		with open(location+"/estimator.pkl", "wb") as f:
			pkl.dump(self.est, f)
	def predict(self, data):
		""" Returns a prediction for the data by querying it to the loaded estimator.

		Parameters
		----------
		data: numpy.ndarray
			Data to be queried to the estimator. Should have format (nDatapoints, nParameters).

		Returns
		-------
		prediction: numpy.ndarray
			Prediction by the loaded estimator for the provided data. Shape of the array depends on the estimator. """
		logger.debug("Querying estimator for prediction (predict)")
		return self.est.predict(data)
	def predict_proba(self, data):
		""" Returns a probability prediction for provided data by querying it to the loaded estimator. Since only classifiers have this functionality in scikit-learn, this method will return `None` if the estimator does not have a prediction_proba method itself.

		Parameters
		----------
		data: numpy.ndarray
			Data to be queried to the estimator. Should have format (nDatapoints, nParameters).

		Returns
		-------
		prediction: numpy.ndarray / `None`
			Prediction by the loaded estimator for the provided data. Shape of the array depends on the estimator. If the loaded estimator does not have a predict_proba method itself, `None` will be returned."""
		logger.debug("Querying estimator for prediction (predict_proba)")
		if hasattr(self.est, predict_proba):			
			return self.est.predict_proba(data)
		logger.debug("Estimator has no predict_proba function")
		return None

class KerasTFEstimator(containers.Estimator):
	""" Interface to a Tensorflow estimator in a Keras wrapper. Inherits its properties from the containers.Estimator class."""
	def __init__(self, path=None, load=False):
		super().__init__(path, load)
		try:
			from keras import __version__ as kerasversion
		except:
			logger.critical("Could not create Estimator based on keras + tensorflow, package keras was not installed.")
			raise exceptions.PhenoAIException("Could not create Estimator based on keras + tensorflow, package keras was not installed.")
		try:
			from tensorflow import __version__ as tfversion
		except:
			logger.critical("Could not create Estimator based on keras + tensorflow, package tensorflow was not installed.")
			raise exceptions.PhenoAIException("Could not create Estimator based on keras + tensorflow, package tensorflow was not installed.")
		self.libaries = {"keras":[kerasversion], "tensorflow":[tfversion]}
	def load(self):
		""" Loads the estimator into the self.est property from the location stored in self.path. """
		from keras.models import load_model
		import tensorflow as tf
		self.est = load_model(self.path+"/estimator.hdf5")
		self.est._make_predict_function()
		self.graph = tf.get_default_graph()
	def save(self, location):
		""" Saves the estimator to provided location.

		Estimator will be stored at specified location as "estimator.hdf5"

		Parameters
		----------
		location: string
			Location where the estimator should be stored, NOT including name and extension of the estimator."""
		if location[-1] == "/":
			location = location[:-1]
		self.est.save(location+"/estimator.pkl")
	def predict(self, data):
		""" Returns a prediction for the data by querying it to the loaded estimator.

		Parameters
		----------
		data: numpy.ndarray
			Data to be queried to the estimator. Should have format (nDatapoints, nParameters).

		Returns
		-------
		prediction: numpy.ndarray
			Prediction by the loaded estimator for the provided data. Shape of the array depends on the estimator. """
		logger.debug("Querying estimator for prediction (predict)")
		with self.graph.as_default():
			return self.est.predict(data)


""" A simple interface to quickly set up and use a Logger instance from the logging module.

This module implements methods to quickly set up a Logger instance with multiple different logging channels. Currently implemented are a stream logger and a file logger.

	import logger
	logger.init(do_stream=True, logpath="output.log")"""

import logging
from logging import Filter
#from logging.handlers import SMTPHandler

__loggername__ = "bsmai_logger"
__streamchannel__ = None
__filechannel__ = None
__streamchannel_defaultlvl__ = logging.INFO
__filechannel_defaultlvl__ = logging.INFO

__colouredstream__ = True
__indentlevel__ = 0
__indentunit__ = "  "


class ColourFilter(Filter):
	""" Adds colour to the output of the stream logging channel if the COLOUR_OUTPUT configuration variable is set to True. """
	def filter(self, record):
		if bool(__colouredstream__):
			if record.levelno == logging.DEBUG:
				record.levellabel = " \033[94mDEBUG\033[0m  "
				record.filteredmessage = record.msg
			elif record.levelno == logging.INFO:
				record.levellabel = "  INFO  "
				record.filteredmessage = record.msg
			elif record.levelno == logging.WARNING:
				record.levellabel = "\033[93mWARNING\033[0m "
				record.filteredmessage = record.msg
			elif record.levelno == logging.ERROR:
				record.levellabel = " \033[91mERROR\033[0m  "
				record.filteredmessage = record.msg
			elif record.levelno == logging.CRITICAL:
				record.levellabel = "\033[1m\033[91mCRITICAL\033[0m\033[0m"
				record.filteredmessage = "\033[91m{}\033[0m".format(record.msg)
		else:
			record.levellabel = "{:^8}".format(record.levelname)
			record.filteredmessage = record.msg
		return record
class IndentFilter(Filter):
	""" Adds indent to the message output of the logging channel. """
	def filter(self, record):
		# Add indenting
		record.msg = "{}{}".format(__indentunit__*__indentlevel__, record.msg)
		return record

def to_stream(lvl=None, colour=True):
	""" Initializes a stream log to the terminal or console
	
	Parameters
	----------
	lvl : integer (default: None)
		Minimum level of the messages to be shown by stream. Uses level integers as defined in the logging package. Will be ignored if do_stream is False. If None, default level will be used as defined in DEFAULT_STREAM_LOGGING_LEVEL in this module.
	colour: boolean (default: True)
		Boolean indicating if stream should use colouring. """
	global __streamchannel__
	add = (__streamchannel__ == None)
	logger = logging.getLogger(__loggername__)
	logger.setLevel(logging.DEBUG)
	if add:
		__streamchannel__ = logging.StreamHandler()
		__streamchannel__.addFilter(IndentFilter())
		__streamchannel__.addFilter(ColourFilter())
	if lvl == None:
		__streamchannel__.setLevel(__streamchannel_defaultlvl__)
	else:
		__streamchannel__.setLevel(lvl)
	if add:
		formatter = logging.Formatter("{asctime:<23} | {levellabel} | {filteredmessage}", style="{")
		__streamchannel__.setFormatter(formatter)
		logger.addHandler(__streamchannel__)

def to_file(logpath, lvl=None):
	""" Initializes a log channel to a file

	Parameters
	----------
	logpath : string
		Location where to log to. File will be created if not already existing.
	lvl : integer (default: None)
		Minimum level of the messages to be stored in the logfile set by logpath. Uses level integers as defined in the logging package. Will be ignored if logpath is None. If not ignored and None, default level will be used as defined in DEFAULT_FILE_LOGGING_LEVEL in this module.
	"""
	global __filechannel__
	add = (__filechannel__ == None)
	logger = logging.getLogger(__loggername__)
	if add:
		__filechannel__ = logging.FileHandler(logpath)
		__filechannel__.addFilter(IndentFilter())
	if lvl == None:
		__filechannel__.setLevel(__filechannel_defaultlvl__)
	else:
		__filechannel__.setLevel(lvl)
	if add:
		formatter = logging.Formatter("{asctime:<23} | {levelname:^8} | {message}", style="{")
		__filechannel__.setFormatter(formatter)
		logger.addHandler(__filechannel__)

def remove_file_channel():
	""" Removes file logging channel from list of logging channels (if one was defined) """
	global __filechannel__
	if not isinstance(__filechannel__, type(None)):
		logger = logging.getLogger(__loggername__)
		logger.removeHandler(__filechannel__)
		__filechannel__ = None

def remove_stream_channel():
	""" Removes stream channel from list of logging channels (if one was defined) """
	global __streamchannel__
	if not isinstance(__streamchannel__, type(None)):
		logger = logging.getLogger(__loggername__)
		logger.removeHandler(__streamchannel__)
		__streamchannel__ = None

def mute(lvl=logging.CRITICAL):
	""" Mutes the output of all logging channels (even those that have nothing to do with PhenoAI) below a specified level.

	Parameters
	----------
	lvl : integer (default: logging.CRITICAL)
		Logging level below which all messages will be muted
	"""
	logging.disable(lvl)
def unmute():
	""" Unmutes the logging module as done through the mute function in this module. """
	logging.disable(logging.NOTSET)
def set_indent(lvl):
	""" Changes the indenting level of the output

	Parameters
	----------
	lvl: integer or string
		Indent level. Can be integer (indicating absolute integer level) or "+" or "-", indicating a relative increase or decrease of indent level with 1. """
	global __indentlevel__
	if isinstance(lvl, int):
		__indentlevel__ = lvl
	elif isinstance(lvl, str):
		if lvl == "+":
			__indentlevel__ += 1
		elif lvl == "-":
			__indentlevel__ -= 1
		else:
			raise Exception("Indent level not recognized")
	elif not isinstance(lvl, type(None)):
		raise Exception("Indent level not recognized")
def colour_stream(colour=True):
	""" Let stream logger use coloured output

	Parameters
	----------
	colour: boolean (default=True)
		Boolean indicating if stream logging should be coloured. """
	global __colouredstream__
	__colouredstream__ = bool(colour)

def debug(message, indent=None):
	""" Send a message to the logger channels with level label logging.DEBUG """
	set_indent(indent)
	logging.getLogger(__loggername__).debug(message)
def info(message, indent=None):
	""" Send a message to the logger channels with level label logging.INFO """
	set_indent(indent)
	logging.getLogger(__loggername__).info(message)
def warning(message, indent=None):
	""" Send a message to the logger channels with level label logging.WARNING """
	set_indent(indent)
	logging.getLogger(__loggername__).warning(message)
def error(message, indent=None):
	""" Send a message to the logger channels with level label logging.ERROR """
	set_indent(indent)
	logging.getLogger(__loggername__).error(message)
def critical(exception, indent=None):
	""" Send a message to the logger channels with level label logging.CRITICAL """
	set_indent(indent)
	logging.getLogger(__loggername__).critical(exception)

if __streamchannel__ == None:
	to_stream()



""" Interface to PhenoAI instances running as server

This client module implements the PhenoAIClient class. The predict method of this class does not perform any prediction itself, but sends the provided data to a by the user configured server that is running an instance of the PhenoAI class in server mode. This server can be the same machine as the one the PhenoAIClient instance is running on. """

from . import splash
splash.splash(False)

import requests
import json
import numpy as np
import os

from . import logger
from . import exceptions
from . import utils
from . import io


class PhenoAIClient(object):
	""" Client class to communicate with instances of PhenoAI running as a server this (or other) machine.

	The PhenoAIClient implements functionality to communicate with other computers on which a PhenoAI instance is running in server mode (this computer can be the same as the one the client is running on). Communication takes place via HTTP requests, but the interface of this class corresponds as much as possible to the one of the PhenoAI class. Before calling the predict method however, the user should first set the server via the constructor (or afterwards via the set_server method).

	Properties
	----------
	address: string
		IP address of the server. 'localhost' is also a valid address.
	ainalysis_ids: list of strings
		List of AInalysis IDs corresponding to the AInalyses to be run at the server side. Be aware that only initialized AInalyses will be selectable via this list.
	port: integer
		Port of the server to which the requests have to be send.

	Constructor arguments
	---------------------
	address: string
		IP address of the server. 'localhost' is also a valid address.
	port: integer
		Port of the server to which the requests have to be send.
	ainalysis_ids: list of strings (default: `None`)
		List of AInalysis IDs corresponding to the AInalyses to be run at the server side. Be aware that only initialized AInalyses will be selectable via this list. """

	def __init__(self, address, port, ainalysis_ids=None):
		self.set_server(address, port)
		self.ainalysis_ids = ainalysis_ids
	def set_server(self, address, port):
		""" Sets the IP address and the port of the server to which the client has to send its requests

		Tries to find the server defined via the supplied address and port. If no connection can be made via check_connection, a ClientException is raised. If a connection could be made the server address and port are saved for all future comminucation.

		Parameters
		----------
		address: string
			IP address of the server. 'localhost' is also a valid address.
		port: integer
			Port of the server to which the requests have to be send. """

		# Check if not only one of the two arguments is attempted to be changed
		if utils.is_none(address) != utils.is_none(port):
			raise exceptions.ClientException("Server address and port have to be set at the same time")
		# Check if arguments are not both None. If so, return (nothing has to be changed)
		if utils.is_none(address):
			return
		# Check if server can be connected
		status = self.check_connection(address, port)
		if not status[0]:
			raise exceptions.ClientException("Client could not connect to server http://{}:{}.\n   {}".format(address, port, status[1]))
		# Set address and port
		self.address = address
		self.port = port
	def check_connection(self, address, port, timeout=5):
		""" Checks if a connection to a server can be made

		Sends a GET request to the server defined via the address and port arguments. If a PhenoAI instance is running on that server that is listening to that port, the returned text will be received by this function and (True, None) will be returned. If an error occured in the request itself or the processing of this request, (False, [errortext]) will be returned, where [errortext] is the description of the error raised.

		Parameters
		----------
		address: string
			IP address of the server. 'localhost' is also a valid address.
		port: integer
			Port of the server to which the requests have to be send.
		timeout: float or `None` (default: 5)
			Time to wait for the server to respond. If set to `None`, script will wait indefinitely.

		Returns
		-------
		can_connect: boolean
			Boolean indicating if a working connection could be established.
		errortext: string
			If can_connect is False, this variable contains the exception that caused the connection to fail. If a connection could be established, this value is None. """

		try:
			if not utils.is_none(timeout):
				response = requests.get("http://{}:{}".format(address, port))
			else:
				response = requests.get("http://{}:{}".format(address, port))
			response = response.text
			if response[:10] == "phenoai-ok":
				return (True, None)
			elif "] Connection refused" in response:
				return (False, "Connection was refused, is there a PhenoAI object running as server on the indicated machine?")
			return (False, response)
		except Exception as e:
			return (False, str(e))
	def predict(self, data, map_data=False, data_ids=None, return_object=True, timeout=5):
		""" Queries the server for prediction on provided data

		Send the provided data to the server set via the set_server method. Depening on the value set for the return_object argument this function will return a copy of the PhenoAIResults object created at the server (True) or a string representation (False). The exact string returned depends on the configuration of the PhenoAI object at the server.

		AInalyses used in prediction at the server depend on the values set in the ainalysis_ids property of this object. Only AInalyses that have been added to the PhenoAI object before run_as_server was called are selectable through this property.

		In contrast to PhenoAI objects not running in server mode, server PhenoAI objects can only process a single file at a time.

		Parameters
		----------
		data: numpy.ndarray or string
			Data can be provided as a numpy ndarray, but it can also be the location of a single data file (e.g. a .slha file). This file will then be read and its complete contents will be sent to the server to be read by the file reader interfaces defined for the running AInalyses. You can only provide a single file per call.
		map_data: boolean or 'both' (default: `False`)
			Boolean indicating if data should be mapped by the AInalyses before running a prediction on it. Can also be 'both', indicating that the prediction has to be run on the mapped and not-mapped data.
		data_ids: List or numpy.ndarray (default: `None`)
			List or numpy array containing IDs for the provided data. Should have same length as data array.
		return_object: boolean (default: `True`)
			Boolean indicating if the function has to return a copy of the PhenoAIResults object created at the server (True) or a string representation of it (False). If a string has to be returned, the contents of this string is determined by the server configuration (see run_as_server method of the phenoai.phenoai class).
		timeout: float or `None` (default: 5)
			Time to wait for the server to respond. If set to `None`, script will wait indefinitely.

		Returns
		-------
		results: PhenoAIResults object or string
			Prediction results created at the server. Type of this value is controlled by the value of the return_object argument of this method. """

		# Create dictionary for the post request
		postdict = {"mapping": 1.0*bool(map_data)}
		# Add ainalysis_ids
		if utils.is_none(self.ainalysis_ids):
			postdict["ainalysis_ids"] = json.dumps("all")
		elif isinstance(self.ainalysis_ids, str):
			postdict["ainalysis_ids"] = json.dumps([self.ainalysis_ids])
		elif isinstance(self.ainalysis_ids, list):
			postdict["ainalysis_ids"] = json.dumps(self.ainalysis_ids)
		elif isinstance(self.ainalysis_ids, float) or isinstance(self.ainalysis_ids, int):
			postdict["ainalysis_ids"] = json.dumps([self.ainalysis_ids])
		else:
			raise exceptions.ClientException("AInalysis IDs provided in ainalysis_ids is not a string, float or int (or list of these elements)")

		# Split different prediction modes
		if isinstance(data, list) or isinstance(data, np.ndarray):
			# Predict labeling from data array
			postdict["mode"] = "values"
			# Check data_ids argument
			if isinstance(data_ids, np.ndarray):
				data_ids = data_ids.tolist()
			if utils.is_none(data_ids):
				data_ids = False
			elif not isinstance(data_ids, list) and not utils.is_none(data_ids):
				raise exceptions.ClientException("Data IDs have to be provided as list or np.ndarray.")
			# Convert numpy.ndarray to list
			if isinstance(data, np.ndarray):
				data = data.tolist()
			# Check if list has correct format (all rows of same length)
			for i in range(len(data)):
				if i != 0:
					if len(data[i]) != L:
						raise exceptions.ClientException("Shape of provided data list/array was inconsistent.")
				else:
					if not isinstance(data[i], list):
						data = [data]
						break
					L = len(data[i])
			# Convert list to json and store in postdict
			postdict["data"] = json.dumps(data)
			postdict["data_ids"] = json.dumps(data_ids)

		elif isinstance(data, str):
			# Predict labeling of file
			postdict["mode"] = "file"
			# Read file to string
			filetext = ''
			with open(data) as file:
				filelines = file.readlines()
				filetext = '\n'.join(filelines)
			if filetext == '':
				raise ClientException("Supplied file '{}' does not have any content.".format(file))
			# Convert filecontents to json
			postdict["data"] = filetext
			postdict["data_ids"] = json.dumps([data])

		else:
			# No prediction could be made
			raise exceptions.ClientException("Provided data has to be a list of lists, numpy.ndarray or the location of a data file. Provided was '{}'.".format(type(data)))

		# Determine return type
		if bool(return_object):
			postdict["get_results_as_string"] = 0.0
		else:
			postdict["get_results_as_string"] = 1.0

		return self.communicate(postdict, return_object, timeout)
	def communicate(self, post_dictionary, return_object=True, timeout=5):
		""" Sends a request to the server

		This method is internally used to make a request to the server. As a user you should use the predict method instead.

		Parameters
		----------
		post_dictionary: dictionary
			Dictionary that has to be send by POST command
		return_object: boolean (default: True)
			Boolean indicating if the function has to return a copy of the PhenoAIResults object created at the server (True) or a string representation of it (False). If a string has to be returned, the contents of this string is determined by the server configuration (see run_as_server method of the phenoai.phenoai class).
		timeout: float or `None` (default: 5)
			Time to wait for the server to respond. If set to `None`, script will wait indefinitely.

		Returns
		-------
		results: PhenoAIResults object or string
			Prediction results created at the server. Type of this value is controlled by the value of the return_object argument of this method. """

		# Read and decode json
		r = requests.post('http://{}:{}'.format(self.address, self.port), data=post_dictionary, timeout=timeout)
		response = r.json()
		# Check if error occured
		if response['status'] == 'error':
			# Error occured, recreate error
			errors = [(name, cls) for name, cls in exceptions.__dict__.items() if isinstance(cls, type)]
			for e in errors:
				if e[0] == response['type']:
					raise e[1]('[@Server] '+response['message'])
			raise Exception('['+e[0]+' @Server]: '+response['message'])
		elif response['status'] == 'ok':
			# No error occured: return results
			if return_object:
				results = io.unpickle(response['results'])
			else:
				results = response["results"]
			# Return data
			return results
		else:
			raise exceptions.ClientException("Response status '{}' was not recognized.".format(response['status']))



""" Implements classes and interfaces used in the handling of individual AInalyses

This module implements the two classes crucial in the handling of AInalyses: AInalysis and AInalysisConfiguration. Where AInalysis implements methods to load relevant AInalysis files and handle data streams to and from its estimator, the AInalysisConfiguration implements checks to validate the configuration of the AInalysis' configuration.yaml file, to prevent errors from happening (and indicate users of possible mistakes).

Although instances of the AInalysis class can be used directly on AInalysis folders, it is in many cases advisable to use instances of phenoai.phenoai as interface instead. """

from . import splash
splash.splash()

import os
import importlib
from importlib.machinery import SourceFileLoader
from inspect import signature
from copy import deepcopy

import numpy as np

from . import __versionnumber__, __version__
from . import containers
from . import estimators
from . import exceptions
from . import io
from . import logger
from . import updatechecker
from . import utils


class AInalysis(object):
	""" Main data juggler class, dealing with dataflows from and to estimators in Estimator instances.

	AInalysis objects form the interface between the user / PhenoAI object and the estimator stored in the AInalysis. Apart from the run method (which starts the AInalysis), it also implements methods to map data and read files. The main purpose of the object is to shift data around in a coherent way (and change it if needed) in such a way that is in agreement with the configuration of the AInalysis in the configuration card.

	Properties
	----------
	id: string
		Unique identifier for this AInalysis. Results from different AInalyses can be distinguished uniquely by this variable, as this value is copied to the AInalysisResults objects created by this AInalysis object.
	configuration: AInalysisConfiguration instance
		The configuration of the AInalysis from the configuration card stored in a AInalysisConfiguration instance.
	estimator: Estimator derived class from the estimators module
		The estimator of the AInalysis stored in an instance of an Estimator class from the estimators module.
	folder: string
		Location of the AInalysis

	Constructor arguments
	---------------------
	folder: string
		Location of the AInalysis. Will be used to load all information from the AInalysis. This location will be stored in the folder property of the object.
	ainalysis_id: string (default=`None`)
		Unique identifier fot his AInalysis. Will set the id property of the object.
	load_estimator: boolean (default=`True`)
		Boolean indicating if the estimator has to be loaded into memory on initialization of the AInalysis object. If set to False the estimator will be loaded on running of the AInalysis. """
	def __init__(self, folder, ainalysis_id = None, load_estimator=True):
		self.ainalysis_id = ainalysis_id
		self.folder = None
		self.estimator = None
		self.configuration = AInalysisConfiguration()
		self.load(folder, load_estimator)
		if utils.is_none(self.ainalysis_id):
			self.ainalysis_id = self.configuration["defaultid"]
	def load(self, folder, load_estimator=True):
		""" Load the configuration from the AInalysis (and the estimator as well, if requested) into memory

		Parameters
		----------
		folder: string
			Location of the AInalysis. Will be used to load all information from the AInalysis. This location will be stored in the folder property of the object.
		load_estimator: boolean (default=`True`)
			Boolean indicating if the estimator has to be loaded into memory. If set to False the estimator will be loaded on running of the AInalysis. """

		logger.info("Loading AInalysis with ID '{}'".format(self.ainalysis_id))
		# Check if provided ainalysis folder exists
		if not os.path.exists(folder):
			raise FileNotFoundError("AInalysis folder '{}' could not be found".format(folder))
		# Alter folder indication if not conform expectation
		if folder[-1] == "/":
			folder = folder[:-1]
		self.folder = folder
		# Load and validate configuration
		self.configuration.load(self.folder+"/configuration.yaml")
		self.configuration.validate()
		# Initialize estimator
		logger.set_indent("+")
		estfac = estimators.EstimatorFactory()
		self.estimator = estfac.create_estimator(self.configuration["class"], self.folder)
		if load_estimator:
			self.estimator.load()
		logger.set_indent("-")
		logger.info("AInalysis loaded".format(self.ainalysis_id))
		self.check_for_update()
	def can_run(self):
		""" Checks if the AInalysis can be run with the information in the AInalysis folder.

		Checks if the configuration has been validated (which it should do automatically on loading via the load method) and if the estimator is loaded. If either of these is not True, the appropriate methods are called to try to fix this (e.g. self.configuration.validate() is called when the configuration is not validated yet. If the check remains untrue after these operations, False is returned. Else True is returned.

		Returns
		-------
		valid: boolean
			Boolean indicating if AInalysis can run with provided information in the AInalysis folder. """
		if not self.configuration.validated:
			self.configuration.validate()
		if not self.estimator.is_loaded():
			self.estimator.load()
		if not self.configuration.validated or not self.estimator.is_loaded():
			return False
		return True
	def read_files(self, paths):
		""" Reads the content from requested files following the definitions provided in the AInalysis configuration.

		Attempts to read the contents of the provided files via the filereader defined in the AInalysis configuration card. If filereader was not set in this configuration or it was set to False (or any equivalent of it), an AInalysisException is raised. Else the defined filereader if used to extract data from the provided files.

		Parameters
		----------
		paths: String or a list of strings
			Locations of the files that should be read.

		Returns
		-------
		data: numpy.ndarray
			A numpy array containing data from the provided files. Shape of the numpy array will be (x, y), where x is the number of provided files and y the number of parameters defined for this AInalysis (as defined in the parameters value in the configuration card). """

		# Check if file reading is enabled
		if utils.is_none(self.configuration["filereader"]) or self.configuration["filereader"] == False:
			raise exceptions.AInalysisException("AInalysis {} has no file reader.".format(self.ainalysis_id))
		# Initialize reader by function if necessary
		if self.configuration["filereader"] == "function":
			functions = SourceFileLoader("module", self.folder+"/functions.py").load_module()
		# Make paths to list if a string
		if isinstance(paths, str):
			paths = [paths]
		# Loop over paths
		logger.debug("AInalysis '{}' is reading {} file(s)".format(self.ainalysis_id, len(paths)))
		data = None
		warning_given = False
		for i in range(len(paths)):
			# Check if file extension is in format list
			if isinstance(self.configuration["filereader.formats"], list) and not warning_given:
				found = False
				for f in self.configuration["filereader.formats"]:
					if path.endswith(f):
						found = True
				if not found:
					logger.warning("One or more files did not have the defined extension for file reading: {}. This might yield errors later in the program.".format(self.configuration["filereader.formats"]))
			# Get data from file
			if self.configuration["filereader"] == "function":
				d = functions.read(paths[i])
			elif isinstance(self.configuration["filereader"], list):
				d = io.read_slha(paths[i], self.configuration["filereader"])
			# Create data array if not existing
			if utils.is_none(data):
				data = np.zeros((len(paths), len(d)))
			# Add new data to data array
			data[i,:] = d
		# Return data to user
		logger.debug("Files read")
		return data
	def map_data(self, data):
		""" Maps provided data

		Maps provided data to the inside of the training region. The mapping method is defined by the mapping variable in the AInalysis configuration card. See the documentation for the AInalysis configuration and the documentation documentation for the mapping procedure for more information.

		This method returns 2 values. The first value (mapped) contains the mapped data, whereas the second (has_changed) is a numpy array containing booleans indicating on a per data point basis if the mapping procedure changed the point. If no mapping procedure was defined in the AInalysis configuration (or it was explicitely disabled) this value will be the plain boolean False.

		Parameters
		----------
		data: numpy.ndarray
			Numpy array of shape (nDatapoints, nParameters) containing the data to be mapped
		
		Returns
		-------
		mapped: numpy.ndarray
			Data after mapping procedure. Has the same shape as the provided data array.
		has_changed: numpy.ndarray
			Numpy array of shape (nDatapoints,) containing boolean values indicating on a per point basis is the data point was changed by the mapping procedure. """

		# If AInalysis does not support mapping return data as is
		logger.debug("AInalysis '{}' is mapping data".format(self.ainalysis_id))
		if (self.configuration["mapping"] == False and isinstance(self.configuration["mapping"], bool)) or utils.is_none(self.configuration["mapping"]):
			logger.debug("Data mapping is not enabled, returning data unaltered")
			return (data, False)
		mapped = deepcopy(data)
		# If mapping is "function" send data to map function and return its result
		if self.configuration["mapping"] == "function":
			logger.debug("Data mapping by function")
			functions = SourceFileLoader("module", self.folder+"/functions.py").load_module()
			mapped = functions.map(mapped, self.configuration["parameters"])
		# If mapping is a floating point, perform SUSY-AI mapping
		elif isinstance(self.configuration["mapping"], float):
			logger.debug("Data mapping by mapping list")
			for i in range(len(self.configuration["parameters"])):
				newmax = (self.configuration["parameters"][i][3] - self.configuration["mapping"]*self.configuration["mapping"])
				newmin = (self.configuration["parameters"][i][2] + self.configuration["mapping"]*self.configuration["mapping"])
				if len(mapped.shape) > 1:
					mapped[mapped[:,i] < newmin, i] = newmin
					mapped[mapped[:,i] > newmax, i] = newmax
				else:
					mapped[i] = newmin
					mapped[i] = newmax
		# Raise exception if mapping mode not understood
		else:
			raise exceptions.AInalysisException("Mapping mode was not recognized. Could not map data.")
		has_changed = (np.sum(np.equal(mapped, data)*1.0, axis=1) != len(self.configuration["parameters"]))
		logger.debug("Data is mapped, returning results")
		return (mapped, has_changed)
	def run(self, data, map_data=False, data_ids=None):
		""" Runs the AInalysis over provided data

		The run method takes data as input and uses the internal estimator to create a prediction for it. Data can be mapped following the mapping procedure for this AInalysis by setting map_data to True. Data and prediction results will automatically be transformed by use of the functions in the functions.py file of this AInalysis before prediction and returning the results to the user respectively.

		Results of the estimation run are returned in an instance of the AInalysisResults class.

		Parameters
		----------
		data: numpy.ndarray, string or list of strings
			Data that has to be subjected to the estimator. Can be the raw data (numpy.ndarray), the location of a file to be read by the built-in file reader in the AInalysis or a list of file locations.
		map_data: boolean (default=False)
			Determines if data has to be mapped before prediction. Mapping uses the map_data method of this AInalysis and follows the mapping procedure defined in the configuration file for this AInalysis.
		data_ids: list or numpy.ndarray (default=`None`)
			List or numpy.ndarray of IDs for the data points. By setting this value, results can be extracted from the AInalysisResults object by using the IDs in this list/array instead of by its location in the result array. If `None`, this functionality will not be available.

		Returns
		-------
		result: AInalysisResults instance
			Instance of the AInalysisResults class containing all prediction results for this AInalysis run. """

		# Read files if file paths were provided as data + do some preprocessing
		logger.info("Running AInalysis '{}'".format(self.ainalysis_id))
		logger.debug("Validating input data type and length")
		if isinstance(data, str):
			data = [data]
		if isinstance(data, list):
			if isinstance(data[0], str):
				data_ids = data 
				data = self.read_files(data)
			else:
				data = np.array(data)
		# Check data shape
		data = np.array(data)
		if len(data.shape) == 1:
			data = data.reshape(1,-1)
		if len(data[0]) != len(self.configuration["parameters"]):
			raise exceptions.AInalysisException("Input data should have {} parameters ({} provided)".format(len(self.configuration["parameters"]), len(data[0])))
		logger.debug("Provided {} data points".format(len(data)))
		# Check if data_ids have same length as data (if set)
		if not utils.is_none(data_ids):
			logger.debug("Validating input data_ids")
			if isinstance(data_ids, list) or isinstance(data_ids, np.ndarray):
				if len(data_ids) != len(data):
					raise exceptions.AInalysisException("Length of provided ID list (data_ids) has to be equal to the length of the provided data array.")
			else:
				raise exceptions.AInalysisException("Data IDs have to be provided as a list or array")
			# Transform entries in data_ids to strings
			if isinstance(data_ids, np.ndarray):
				data_ids = data_ids.tolist()
			for i in range(len(data_ids)):
				data_ids[i] = str(data_ids[i])
			logger.debug("Data IDs validated")
		# Check if AInalysis is ready for run
		estimator_was_loaded = self.estimator.is_loaded()
		if not self.can_run():
			raise exceptions.AInalysisException("Cannot run AInalysis {}".format(self.ainalysis_id))
		# Map data if requested
		if map_data:
			logger.info("Mapping data")
			data, mapped = self.map_data(data)
		else:
			mapped = False
		# Create result object
		result = containers.AInalysisResults(self.ainalysis_id, self.configuration, data, data_ids, mapped)
		# Perform data transformation
		functions = SourceFileLoader("module", self.folder+"/functions.py").load_module()
		if "transform" in dir(functions):
			logger.debug("Transforming data")
			data = functions.transform(data)
		# Perform prediction
		logger.info("Perform prediction")
		logger.set_indent("+")
		predictions = self.estimator.predict(data)
		logger.set_indent("-")
		# Perform inverse transformation on prediction results
		if "transform_predictions" in dir(functions):
			logger.debug("Transforming results")
			predictions = functions.transform_predictions(predictions)
		# Store results in AInalysisResults
		result.predictions = predictions
		# Remove estimator from memory if was not loaded
		if not estimator_was_loaded:
			logger.debug("Clearing estimator from memory")
			self.estimator.clear()
		# Return result object
		logger.info("Prediction finished, result returned")
		return result
	def check_for_update(self, print_info=True):
		""" Checks for update of the AInalysis

		Uses the updatechecker.check_ainalysis_update function to check if there is an updated version of this AInalysis available. For this the unique_db_id entry in the AInalysis configuration card is used. If an update is found, this is communicated to the user via the logger module.

		Parameters
		----------
		print_info: boolean (default=True)
			Boolean indicating if information should be printed to terminal

		Returns
		-------
		error: boolean
			Boolean indicating if an error occurred on the server during processing of request. If True, return variable `text` will be the unformatted error message.
		update: boolean
			Boolean indicating if an update is available. If return variable `error` is True, this variable will be `None`.
		text: string
			Return message. If return variable `update` is True, this message will be formatted. """

		if utils.is_none(self.configuration["uniquedbid"]):
			logger.info("No uniquedbid defined in configuration, no update checking will be performed")
			return (True, None, "no uniquedbid defined in configuration, no update checking will be performed")

		error, updatable, txt = updatechecker.check_ainalysis_update(self.configuration["unique_db_id"], self.configuration["ainalysisversion"])
		if print_info:
			print("\n===========================================================================\n|       Always look up if this AInalysis is applicable to your case       |\n|                     Don't forget to cite the author                     |\n===========================================================================\n")
			if not error:
				print(txt+"\n")
		else:
			logger.warning("Could not check for updates of this ainalysis, because: "+txt)
		return (error, updatable, txt)


class AInalysisConfiguration(containers.Configuration):
	""" Interface and validation class for the AInalysis configuraton.

	Interface class that inherits from containers.Configuration. This class is meant to contain the information from AInalysis configuration cards in its configuration property (see containers.Configuration). This derived class builts onto its parent class by implementing validation methods (of which `validate` is the main one). It adds no new properties.

	Any changes made by the validation methods to the configuration are made to the configuration in active memory. Changes are never made to the configuration file itself. """
	def validate(self, validate_checksum=True):
		""" Validates the configuration in the configuration property

		Runs all relevant validation methods of the AInalysisConfiguration class on the loaded configuration.

		Parameters
		----------
		validate_checksum: boolean (default: True)
			Indicates if the checksum in the folder has to be validated as well.

		Returns
		-------
		valid: boolean
			Boolean value indicating if the configuration was good as-is. If any value in the configuration file has to be altered or new ones had to be added, this value will be False. Else it will be True"""

		logger.info("Validating AInalysis configuration")
		valid = True
		# General information
		valid *= self.validate_uniquedbid()
		valid *= self.validate_defaultid()
		valid *= self.validate_version()
		valid *= self.validate_phenoaiversion()
		valid *= self.validate_class()
		valid *= self.validate_libaries()
		valid *= self.validate_type()
		valid *= self.validate_output_and_classes()
		if validate_checksum:
			valid *= self.validate_checksum()
		# Parameters
		valid *= self.validate_parameters()
		# Estimator type specifics
		if self.configuration["type"] == "classifier":
			valid *= self.validate_classifier_calibrated()
			if self.configuration["classifier.calibrated"] == False:
				valid *= self.validate_classifier_calibrate()
				valid *= self.validate_classifier_calibratedata()
		elif self.configuration["type"] == "regressor":
			pass
		# File reading
		valid *= self.validate_filereader()
		# Mapping
		valid *= self.validate_mapping()
		self.validated = True
		if valid:
			logger.info("Configuration valid")
		else:
			logger.warning("Configuration did not fully specify AInalysis behaviour: some behaviour might be infered / set to default")
		return bool(valid)

	# General information validator
	def validate_uniquedbid(self):
		""" Checks for the presence of `unique_db_id` in the AInalysis configuration

		Checks if there is an entry in the configuration named unique_db_id. If not present, this entry will be made (set to `None`). Not providing this value in the configuration does not mean the configuration is invalid, so the returned boolean will always be True (in order to let all validate_* methods have the same interface and behaviour).

		Returns
		-------
		valid: boolean
			Will always be True. """
		if "uniquedbid" not in self.configuration:
			self.configuration["uniquedbid"] = None
			logger.debug("Configuration entry 'uniquedbid' not found; entry is set to `None`")
		return True
	def validate_defaultid(self):
		""" Checks if a default ID (`defaultid`) was provided in the AInalysis configuration

		Checks if there is an entry in the configuration named defaultid. If not present, this entry will be made and set to a random 5 character long string. False will be returned if this happens. In all other cases True will be returned.

		Returns
		-------
		valid: boolean
			Boolean indicating if defaultid was set in the configuration."""
		if "defaultid" not in self.configuration:
			logger.warning("No default id was found. Will create a random string to serve as one.")
			defaultid = utils.random_string(5)
			self.configuration["defaultid"] = defaultid
			return False
		logger.debug("Configuration entry 'defaultid' was validly defined.")
		return True
	def validate_version(self):
		""" Checks for the presence of AInalysis version information (`ainalysisversion`) in the AInalysis configuration.

		Checks if the version number of the AInalysis was provided in the AInalysis configuration. If no version number was provided, the `ainalysisversion` entry will be made in the configuration and set to `1`. This method will always return True, since the configuration does not get less valid by not providing this value.

		Returns
		-------
		valid: boolean
			Will always be True. """
		if "ainalysisversion" not in self.configuration:
			logger.warning("No version number provided for this AInalysis. No checking for updates can be performed.")
			self.configuration["ainalysisversion"] = 1
		else:
			logger.debug("Configuration entry 'ainalysisversion' was validly defined.")
		return True
	def validate_phenoaiversion(self):
		""" Checks if the currently running PhenoAI version (`phenoaiversion`) is actively supported by this AInalysis.

		Checks if the configuration contains an entry `phenoaiversion`. If so the version defined in the AInalysis configuration is checked against the versionnumber of the currently running PhenoAI package. If the configuration explicitely mentions the current PhenoAI version, True is returned. In all other cases False is returned.

		Returns
		-------
		valid: boolean
			Boolean value indicating if currently running PhenoAI versionnumber is mentioned in the AInalysis configuration as actively supported/tested. """
		if "phenoaiversion" not in self.configuration:
			logger.warning("No supported PhenoAI version was defined. Will assume currently used version is supported.")
			return False
		if isinstance(self.configuration["phenoaiversion"], int):
			self.configuration["phenoaiversion"] = [self.configuration["phenoaiversion"]]
		if __version__ not in self.configuration["phenoaiversion"]:
			logger.warning("This AInalysis does not explicitely support the current version of PhenoAI. Program will continue but errors may occur.")
			return False
		logger.debug("Configuration entry 'phenoaiversion' was validly defined.")
		return True
	def validate_class(self):
		""" Checks if the estimator class (`class`) is supported.

		Checks if `class` is defined in the configuration. If so, the value for this entry is checked for support by the current version of PhenoAI. Currently supported values are 'sklearnestimator' and 'kerastfestimator'. Only if the value in the configuration is one of these values, True will be returned. In all other cases a ConfigurationException will be raised. 

		Returns
		-------
		valid: boolean
			Boolean value indicating if the `class` entry exists in the AInalysis configuration and if it is supported by the current version of PhenoAI."""
		if "class" not in self.configuration:
			logger.error("No class was provided for the AInalysis estimator.")
			raise exceptions.ConfigurationException("No class was provided for the AInalysis estimator.")
			return False
		self.configuration["class"] = self.configuration["class"].lower()
		if self.configuration["class"] not in ["sklearnestimator", "kerastfestimator"]:
			logger.error("Class '{}' is not implemented in this version of PhenoAI.".format(self.configuration["class"]))
			raise exceptions.ConfigurationException("Class '{}' is not implemented in this version of PhenoAI.".format(self.configuration["class"]))
			return False
		logger.debug("Configuration entry 'class' was validly defined.")
		return True
	def validate_libaries(self):
		""" Checks if the needed libraries defined in the configuration (`libraries`) are installed with supported versions.

		Checks if the configuration contains a `libraries` entry and loops over all entries in it if it is found. For each of these libraries the version numbers mentioned in the configuration are checked against the version of the library currently installed. If all libraries defined in the configuration are installed with supported versions, True is returned. In all other cases False is returned.

		Returns
		-------
		valid: boolean
			Boolean False if there are required libraries installed with version numbers that are not actively tested / supported. True if all installed and required libraries have the correct version. """

		if "libraries" not in self.configuration:
			logger.warning("No information was provided on needed libraries. This might cause problems during prediction.")
			return False
		# Load libraries and save versions for installed ones
		libraries = {"sklearn": None, "keras": None, "tensorflow": None}
		for lib in libraries:
			try:
				package=importlib.import_module(lib)
				libraries[lib] = package.__version__
			except:
				pass
		unsupported = 0
		for lib in self.configuration["libraries"]:
			if utils.is_none(self.configuration["libraries"][lib]):
				continue
			if libraries[lib] == None:
				logger.error("AInalysis uses library '{}', but it is not installed.".format(lib))
				raise exceptions.AInalysisException("AInalysis uses library '{}', but it is not installed.".format(lib))
				return False
			elif libraries[lib] not in self.configuration["libraries"][lib]:
				logger.warning("Explicitly supported versions of {} don't list the currently installed one ({}). This might cause errors down the line.".format(lib, libraries[lib]))
		if unsupported == 0:
			logger.debug("Configuration entry 'libraries' was validly defined.")
		return (unsupported == 0)
	def validate_type(self):
		""" Checks if the estimator type (`type`) was validly defined in the configuration.

		Checks if an entry `type` was defined in the AInalysis configuration and if it equals 'classifier' or 'regressor'. If so, True is returned. In all other cases a ConfigurationException is raised.

		Returns
		-------
		valid: boolean
			Boolean value indicating if the estimator type (`type` in the configuration) was defined as 'classifier' or 'regressor'. """

		if "type" not in self.configuration:
			logger.error("No estimator type was defined.")
			raise exceptions.ConfigurationException("No estimator type was defined.")
			return False
		if self.configuration["type"] not in ["classifier", "regressor"]:
			logger.error("Estimator type '{}' was not recognized.".format(self.configuration["type"]))
			raise exceptions.ConfigurationException("Estimator type '{}' was not recognized.".format(self.configuration["type"]))
			return False
		logger.debug("Configuration entry 'type' was validly defined.")
		return True
	def validate_output_and_classes(self):
		""" Checks if output value name(s) (`output`) is(are) defined in the AInalysis configuration.

		Checks if there is an entry named `output` in the AInalysis configuration. If not, a ConfigurationException is raised, else True is returned. If the value for this entry is not a list, the entry is converted to a list containing the provided entry. 

		Returns
		-------
		valid: boolean
			Boolean value indicating if `output` was defined in the AInalysis configuration."""
		if "output" not in self.configuration:
			logger.error("No information on estimator output is provided.")
			raise exceptions.ConfigurationException("No information on estimator output is provided.")
			return False
		if not isinstance(self.configuration["output"], str):
			logger.error("Estimator output should have been defined as a string.")
			raise exceptions.ConfigurationException("Estimator output should have been defined as a string.")
			return False

		if self.configuration["type"] == "classifier":
			if "classes" not in self.configuration:
				logger.error("No information on estimator classes is provided.")
				raise exceptions.ConfigurationException("No information on estimator output is provided.")
				return False
			if not isinstance(self.configuration["classes"], dict):
				logger.error("Estimator classes should have been defined as a dictionary.")
				raise exceptions.ConfigurationException("Estimator classes should have been defined as a dictionary.")
				return False
			classes = {}
			for k,v in self.configuration["classes"].items():
				try:
					classes[int(k)] = v
				except:
					logger.error("Classes dictionary should contain integers as keys.")
					raise exceptions.ConfigurationException("Classes dictionary should contain integers as keys.")
					break
			self.configuration["classes"] = classes
			logger.debug("Configuration entry 'classes' was validly defined.")

		logger.debug("Configuration entry 'output' was validly defined.")
		return True
	def validate_checksum(self):
		""" Checks the validness of the AInalysis Checksum.

		Checks if a checksum is stored in the AInalysis folder and calculates if this checksum corresponds to the one calculated from the AInalysis folder as it currently exists. If the checksum either does not exist or any of the calculated checksums does not correspond to the one stored, False is returned. In any other case True is returned. """
		try:
			stored_checksums = io.read_checksum(self.folder+"/checksums.sfv")
		except exceptions.FileIOException:
			logger.warning("No checksum file found. Could not check if data is uncorrupted.")
			return False
		calculated_checksums = utils.calculate_ainalysis_checksums(self.folder)
		invalid_checksums = []
		# Check checksums of important files
		for i in calculated_checksums:
			if i not in stored_checksums:
				logger.warning("No checksum found for '{}' in the checksums.sfv file, could not check if data is uncorrupted.".format(i))
				return False
			if calculated_checksums[i] != stored_checksums[i]:
				invalid_checksums.append(i)
		# Check if checksums were invalid and if so show error (but not fatal)
		if len(invalid_checksums) > 0:
			logger.warning("Checksum of following files where invalid:")
			logger.set_indent("+")
			for i in range(len(invalid_checksums)):
				logger.warning("- {}".format(invalid_checksums[i]))
			logger.warning("This hints at corruption of data (or at the very least: the data has been changed). Use with care and if possible redownload the AInalysis.", "-")
			return False
		if len(invalid_checksums) == 0:
			logger.debug("Checksums are valid.")
		return True
	def validate_classifier_calibrated(self):
		""" Check if information was provided in the configuration on calibration of the estimator.

		Checks if there is an estimator type defined in the configuration (`type`) and if it equals 'classifier'. If so, the presence of `classifier.calibrated` is checked. If it is not present in the configuration False is returned. In all other cases True is returned.

		Returns
		-------
		valid: boolean
			If the estimator is a classifier and calibration information is *not* provided in the AInalysis configuration, this value is False. In all other cases this value is True. """
		if "type" in self.configuration and self.configuration["type"] == "classifier":
			if "classifier.calibrated" not in self.configuration:
				self.configuration["classifier.calibrated"] = False
				logger.warning("No calibration information was provided. Assuming classifier is not calibrated.")
				return False
			self.configuration["classifier.calibrated"] = bool(self.configuration["classifier.calibrated"])
			logger.debug("Configuration entry 'classifier.calibrated' was validly defined.")
		return True
	def validate_classifier_calibrate(self):
		""" Checks if information on calibration of estimator output is provided.

		Checks if the estimator is a classifier. If it is and the classifier is not calibrated, the presence of `classifier.calibrate` is checked. If this entry is not provided, False is returned. In all other cases True is returned.

		If `classifier.calibrate` is indeed found, it is cast to a boolean to avoid problems in other parts of the program. If no `calssifier.calibrate` was found but all requirements (uncalibrated classifier) were fulfilled, the entry is set to False.

		Returns
		-------
		valid: boolean
			If the estimator is an uncalibrated classifier and no information was provided in the `classifier.calibrate` entry of the configuration, this value is False. In all other cases it is False. """
		if "type" in self.configuration and self.configuration["type"] == "classifier":
			if "classifier.calibrated" in self.configuration and self.configuration["classifier.calibrated"] == False:
				if "classifier.calibrate" not in self.configuration:
					self.configuration["classifier.calibrate"] = False
					logger.warning("No information provided on whether or not to calibrate classifier output. Assuming this is not the case.")
					return False
				self.configuration["classifier.calibrate"] = bool(self.configuration["classifier.calibrate"])
			logger.debug("Configuration entry 'classifier.calibrate' was validly defined.")
		return True
	def validate_classifier_calibratedata(self):
		""" Checks if information provided to calibrate estimator output is formatted correctly.

		Checks if the estimator is a classifier (`type`) that is uncalibrated (`classifier.calibrated`) and of which the output has to be calibrated (`classifier.calibrate`). If one of these requirements is not met, True is returned. If all requirements are met, the presence of `classifier.calibrate.bins` and `classifier.calibrate.values` is checked. The values of these entries in the configuration have to be lists with equal length. If the entries do not exist or the lists of equal length is not met, a ConfigurationException is raised. In all other cases, True is returned.

		If values and bins were validly defined, they are converted from list to numpy array.

		Returns
		-------
		valid: boolean
			True if information on classifier output calibration was provided in correct format (see above). If this is not the case, this value is False. In all other cases this value is True. """
		if "type" not in self.configuration or self.configuration["type"] != "classifier":
			return True
		if "classifier.calibrated" not in self.configuration or self.configuration["classifier.calibrated"] == True:
			return True
		if "classifier.calibrate" not in self.configuration or self.configuration["classifier.calibrate"] == False:
			return True
		if "classifier.calibrate.bins" not in self.configuration:
			self.configuration["classifier.calibrate"] = False
			logger.error("No information on calibration bins is provided. No calibration will take place.")
			raise exceptions.ConfigurationException("No information on calibration bins is provided. No calibration will take place.")
			return False
		if "classifier.calibrate.values" not in self.configuration:
			self.configuration["classifier.calibrate"] = False
			logger.warning("No information on calibration values is provided. No calibration will take place.")
			raise exceptions.ConfigurationException("No information on calibration values is provided. No calibration will take place.")
			return False
		if not isinstance(self.configuration["classifier.calibrate.bins"], list) and not isinstance(self.configuration["classifier.calibrate.bins"], np.ndarray):
			self.configuration["classifier.calibrate"] = False
			logger.error("Calibration bin information is not provided as a list. No calibration will take place.")
			raise exceptions.ConfigurationException("Calibration bin information is not provided as a list. No calibration will take place.")
			return False
		if not isinstance(self.configuration["classifier.calibrate.values"], list) and not isinstance(self.configuration["classifier.calibrate.values"], np.ndarray):
			self.configuration["classifier.calibrate"] = False
			logger.error("Calibration values information is not provided as a list. No calibration will take place.")
			raise exceptions.ConfigurationException("Calibration values information is not provided as a list. No calibration will take place.")
			return False
		if len(self.configuration["classifier.calibrate.values"]) != len(self.configuration["classifier.calibrate.bins"]):
			self.configuration["classifier.calibrate"] = False
			logger.error("Number of calibration bins and values do not correspond. No calibration will take place.")
			raise exceptions.ConfigurationException("Number of calibration bins and values do not correspond. No calibration will take place.")
			return False
		self.configuration["classifier.calibrate.values"] = np.array(self.configuration["classifier.calibrate.values"])
		self.configuration["classifier.calibrate.bins"] = np.array(self.configuration["classifier.calibrate.bins"])
		logger.debug("Configuration entry 'classifier.calibrate.bins' was validly defined.")
		logger.debug("Configuration entry 'classifier.calibrate.values' was validly defined.")
		return True
	def validate_filereader(self):
		""" Checks if a valid filereader was provided in the AInalysis configuration.

		Checks if `filereader` was provided in the AInalysis configuration. If not, False is returned. If it is provided, it has to be "function", a list or `None`, otherwise False is returned. If the entry is set to "function", the `data` function has to be defined in the "functions.py" file in the AInalysis folder, which takes 1 input argument (file location). If this function was not found or the function takes a different kind of parameters False is returned.

		If the `filereader` entry is a list, it has to be a list of lists of length 2 following the format [BLOCK, SWTICH], indicating the blocks and switches from the entries in a .slha file from which the information has to be extracted. If this format is violated, False is returned.

		If False is returned, the `filereader` entry is set to None.

		In all other cases True is returned.

		Returns
		-------
		valid: boolean
			True if a valid filereader was defined (see above). If not, this value is False. """
		if "filereader" not in self.configuration:
			self.configuration["filereader"] = None
			logger.warning("No readmode defined.")
			return False
		if isinstance(self.configuration["filereader"], bool) and not self.configuration["filereader"]:
			self.configuration["filereader"] = None
		if self.configuration["filereader"] == "function":
			try:
				functions = SourceFileLoader("module", self.folder+"/functions.py").load_module()
				narguments = len(signature(functions.read).parameters)
				if narguments != 1:
					self.configuration["filereader"] = None
					logger.warning("Function for data reading 'read' takes {} arguments, 1 argument expected. No automated file reading can take place.".format(narguments))
					return False
			except FileNotFoundError:
				# File functions.py was not found in AInalysis folder
				self.configuration["filereader"] = None
				logger.warning("No functions file could be found in AInalysis folder. No automated file reading can take place.")
				return False
			except AttributeError:
				# No function defined called data
				self.configuration["filereader"] = None
				logger.warning("No reader function 'read' defined in functions file. No automated file reading can take place.")
				return False
		elif self.configuration["filereader"] == "None":
			self.configuration["filereader"] = None
		elif isinstance(self.configuration["filereader"], list):
			for i in range(len(self.configuration["filereader"])):
				if len(self.configuration["filereader"][i]) != 2:
					self.configuration["filereader"] = None
					logger.warning(".slha readerlist has an invalid format. No automated file reading can take place")
					return False
			# Check if length of reader list is equal to the number of parameters
			if len(self.configuration["filereader"]) != len(self.configuration["parameters"]):
				self.configuration["filereader"] = None
				logger.warning("Length of .slha readerlist should correspond to the length of the parameter list.")
				return False
		elif not utils.is_none(self.configuration["filereader"]):
			logger.warning("Readmode '{}' was not recognized. No reading can take place.".format(self.configuration["filereader"]))
			self.configuration["filereader"] = None
			return False
		logger.debug("Configuration entry 'filereader' was validly defined.")
		
		# validate filereader formats
		if "filereader.formats" not in self.configuration:
			self.configuration["filereader.formats"] = None
		if utils.is_none(self.configuration["filereader.formats"]):
			logger.warning("Filereader.formats not defined / defined as None, unclear which file formats can be read.")
			return False
		elif isinstance(self.configuration["filereader.formats"], list):
			formats = self.configuration["filereader.formats"]
			for i in range(len(formats)):
				if not isinstance(formats[i], str):
					formats[i] = None
					logger.warning("Filereader.formats ill defined, should be a string or list of strings.")
			for i in range(len(formats)-1,-1,-1):
				if utils.is_none(formats[i]):
					del(formats[i])
			if len(formats) == 0:
				logger.warning("No valid formats defined, no file format validation will be performed.")
				self.configuration["filereader.formats"] = None
				return False
			self.configuration["filereader.formats"] = formats
		elif isinstance(self.configuration["filereader.formats"], str):
			self.configuration["filereader.formats"] = [self.configuration["filereader.formats"]]
		logger.debug("Configuration entry 'filereader.formats' was validly defined.")
		return True
	def validate_mapping(self):
		""" Checks if information on data mapping is provided in the configuration.

		Checks for the presence of `mapping` in the AInalysis configuration. If it is present, but is not of type boolean, type float or equals "function", False is returned. If it equals "function", but no valid `map` function is defined in the functions.py file in the AInalysis folder (valid means: existing and taking 1 argument, namely the data array), False is returned. In all other cases True is returned.

		If the entry is the boolean True it is changed to 0.0. This does not change the functionality of the code, but makes interpretation of the entry easier. If no entry was found `mapping` is set to False.

		Returns
		-------
		valid: boolean
			True is `mapping` is set in the configuration and is of valid format (boolean, float or equalling "function"). In all other cases this value is False. """
		if "mapping" not in self.configuration:
			self.configuration["mapping"] = False
			logger.warning("No information was provided if mapping should take place. Assuming not.")
			return False
		if not isinstance(self.configuration["mapping"], bool) and self.configuration["mapping"]!="function" and not isinstance(self.configuration["mapping"], float):
			self.configuration["mapping"] = False
			logger.warning("Mapping mode was not recognized. Has to be False, 'function' or a floating point number. Mapping is disabled for this AInalysis.")
			return False
		if self.configuration["mapping"] == "function":
			try:
				functions = SourceFileLoader("module", self.folder+"/functions.py").load_module()
				narguments = len(signature(functions.map).parameters)
				if narguments != 1:
					self.configuration["mapping"] = False
					logger.warning("Function for data mapping 'map' takes {} arguments, 1 argument expected. No data mapping can take place.".format(narguments))
					return False
			except FileNotFoundError:
				# File functions.py was not found in AInalysis folder
				self.configuration["mapping"] = False
				logger.warning("No functions file could be found in AInalysis folder. No data mapping can take place.")
				return False
			except AttributeError:
				# No function defined called data
				self.configuration["mapping"] = False
				logger.warning("No 'map' function defined in functions file. No data mapping can take place.")
				return False
		if self.configuration["mapping"] == True and isinstance(self.configuration["mapping"], bool):
			self.configuration["mapping"] = 0.0
		logger.debug("Configuration entry 'mapping' was validly defined.")
		return True
	def validate_parameters(self):
		""" Checks if information on input parameters is provided in the AInalysis configuration.

		Checks if `parameters` is set in the AInalysis configuration. If not, a ConfigurationException is raised. If it is set, it has to be a list consisting out of lists of length 3 with entries [NAME, MINIMUM, MAXIMUM]. If this format is broken in *any* of the list entries, a ConfigurationException is raised. In all other cases True is returned.

		Returns
		-------
		valid: boolean
			True is `parameters` entry in the AInalysis configuration was set and follows the required format (see above). If not, this value equals False. """
		if "parameters" not in self.configuration:
			logger.error("No parameter information was provided.")
			raise exceptions.ConfigurationException("No parameter information was provided.")
			return False
		for i in range(len(self.configuration["parameters"])):
			if not isinstance(self.configuration["parameters"][i], list) or len(self.configuration["parameters"][i]) != 4:
				logger.error("Parameter definitions dont follow the [NAME, UNIT, MIN, MAX] format.")
				raise exceptions.ConfigurationException("Parameter definitions dont follow the [NAME, UNIT, MIN, MAX] format.")
				return False
		logger.debug("Configuration entry 'parameters' was validly defined.")
		return True




""" This module implements all custom exceptions that can be raised by the PhenoAI package. """

class FileIOException(Exception):
	""" Exception class of which instances can be raised by functions in the io module. """
	pass
class ConfigurationException(Exception):
	""" Exception class of which instances can be raised by classes.Configuration derived classes. """
	pass
class AInalysisException(Exception):
	""" Exception class of which instances can be raised by the ainalyses.AInalysis class. """
	pass
class PhenoAIException(Exception):
	""" Exception class of which instances can be raised by the phenoai.phenoai class. """
	pass
class ResultsException(Exception):
	""" Exceptions class of which instances can be raised by the phenoai.PhenoAIResults and ainalyses.AInalysisResults classes. """
	pass
class ServerException(Exception):
	""" Exception class of which instances can be raised by the phenoai.phenoaiServer class. """
	pass
class ClientException(Exception):
	""" Exception class of which instances can be raised by the client module. """
	pass
class MakerError(Exception):
	""" Exception class of which instances can be raised by the maker module. """


""" The PhenoAI module implements the main interface for the PhenoAI package: the PhenoAI class.

This module contains the PhenoAI class: the primary way to interact with the package, which implements methods to handle AInalyses, run predictions in them and run the package as a server. """

from . import splash
splash.splash()

import traceback
import ast
import cgi
import os
import sys
from http.server import BaseHTTPRequestHandler, HTTPServer
from socketserver import ThreadingMixIn

import numpy as np
import json
import pyslha

from . import ainalyses
from . import containers
from . import client
from . import exceptions
from . import io
from . import logger
from . import utils

__serverinstance__ = None

class PhenoAI(object):
	""" The main interface for running BSM-AI.

	The PhenoAI class provides an interface to send data to (multiple) AInalyses and package their results in an instance of PhenoAIResults. Its main method is the run method, which calls the run method of each AInalyses stored in the PhenoAI instance.

	Instances can run in one of two modes: dynamic or static. In static mode, the estimator of each of the AInalyses is loaded into memory and they will stay there, even after prediction. This makes making a new prediction very fast, since the estimator has not to be loaded again. However, they do take up memory space, which can be a problem in low-memory environments or when running multiple AInalyses in the same PhenoAI instance. Dynamic mode solves this problem by only loading the estimator of an AInalysis if it is needed to make a prediction. If it is loaded after this check and prediction has finished, the estimator is cleared from memory. This of course comes at the cost of speed: loading the estimator takes time.

	Instances of this class can also be run as a server via the run_as_server method. This locks the PhenoAI instance in its current modus, but allows client.phenoaiClient instances to communicate with it over the network or internet.

	Properties
	----------
	ainalyses: list of ainalyses.AInalysis instances
		List of AInalyses managed by the PhenoAI instance. These instances can be queried for prediction on data. AInalyses can be added via the add method.
	dynamic: boolean
		If True, estimators of the AInalyses will be loaded only when necessary. When finished, the estimator will be cleared from memory. This property is useful when running on low RAM machines or when a large collection of AInalyses is stored in the PhenoAI instance.

	Constructor arguments
	---------------------
	dynamic: boolean (default: True)
		If True, estimators of the AInalyses will be loaded only when necessary. When finished, the estimator will be cleared from memory. This property is useful when running on low RAM machines or when a large collection of AInalyses is stored in the PhenoAI instance. """

	def __init__(self, dynamic = True):
		self.ainalyses = []
		self.dynamic = dynamic
	def add(self, ainalysis_folder, ainalysis_id = None):
		""" Adds an AInalysis to the PhenoAI instance

		Adds and AInalysis to the list stored in the ainalyses property of this PhenoAI instance. This allows the PhenoAI instance to query it for its prediction on provided data. This method will raise a PhenoAIException if the provided AInalysis ID would not be unique in the list of AInalyses already stored in the PhenoAI instance.

		Parameters
		----------
		ainalysis_folder: string
			Path to the AInalysis folder
		ainalysis_id: string or `None` (default: `None`)
			ID of the AInalysis to be used in this instance of phenoai. This ID will also be used as identifier for the created AInalysisResults objects by that AInalysis. If set to `None` the default ID defined in the AInalysis configuration will be used. """
		logger.info("Adding AInalysis to PhenoAI object")
		logger.set_indent("+")
		a = ainalyses.AInalysis(ainalysis_folder, ainalysis_id, not self.dynamic)
		logger.set_indent("-")
		if not utils.is_none(self.get(a.ainalysis_id)):
			aid = a.ainalysis_id
			del a
			raise exceptions.phenoaiException("Cannot add AInalysis '{}' with id '{}', ID is already known".format(ainalysis_folder, aid))
		logger.info("AInalysis '{}' added to PhenoAI object".format(a.ainalysis_id))
		self.ainalyses.append(a)
	def get(self, ainalysis_id):
		""" Returns the AInalysis instance with provided AInalysis ID

		Parameters
		----------
		ainalysis_id: string
			ID of the AInalysis that has to be returned.

		Returns
		-------
		ainalysis: AInalysis instance or `None`
			AInalysis with corresponding ID. If no such AInalysis was found, `None` is returned."""
		for ainalysis in self.ainalyses:
			if ainalysis.ainalysis_id == ainalysis_id:
				return ainalysis
		return None
	def is_dynamic(self, mode=True):
		""" Sets how memory needed for estimators has to be allocated

		This method determines how memory needed for the estimators in the AInalyses is allocated and freed. There are two modes: dynamic (True) and static (False). In static mode all estimators are loaded into memory and stay there, even after prediction. In dynamic mode estimators are only loaded when needed and cleared from memory afterwards.

		Parameters
		----------
		mode: boolean
			Boolean indicating which mode the PhenoAI instance has to run in. Can either be True (dynamic) or False (static)."""
		self.dynamic = mode
		if mode:
			logger.info("PhenoAI run mode: dynamic")
		else:
			logger.info("PhenoAI run mode: static")
		for a in self.ainalyses:
			if mode:
				a.estimator.clear()
			else:
				a.estimator.load()
	def run_as_server(self, address, port, logging_path=None, to_string_function=None):
		""" Lets the PhenoAI instance into a server, allowing it to perform predictions on data sent to it from an external script.

		Calling this function lets the PhenoAI instance listen to prediction requests coming in via a specific port and to a specific address instead of prediction requests coming in from the rest of the script. As a consequence, calling this function includes an infinite loop to facilitate the listening for requests. For safety reasons the port this instance will listen to has to be larger then 1024.

		The requests coming in should be HTTP requests and can be created via the client.phenoaiClient class. Important to note is that these requests can be made on the same computer as this PhenoAI instance is running, but also on another computer that is connected to this one via a network (e.g. the internet).

		The split between the client and the server makes it possible to let the client script be written in any programming language, as long as the request follows the correct format. However, the server will return, by default, an instance of the PhenoAIResults class, which can only be used in Python. This problem is solved by letting the server extract (and calculate) the relevant information from the PhenoAIResults object and letting it return not this object, but the relevant information in string form. Since the question "what is the relevant information?" will be answers differently for each use-case, the post-processing will be done by a function written by the user and supplied to this run_as_server method via the to_string_function argument. This function should take the PhenoAIResults instance as its only argument. For example:

				phenoai = PhenoAI()
				def stringify_results(PhenoAIResults):
					return ",".join(PhenoAIResults.predictions)
				phenoai.run_as_server("localhost", 1992, to_string_function = stringify_results)

		The default stringify_results function returns the predictions in .csv format, where the first line contains the file IDs (if provided) and all subsequent lines the predictions for each of the AInalyses for each for each of the data points.

		Whether or not the server returns the full PhenoAIResults object or the generated string is always determined by the client via the requests made.

		Parameters
		----------
		address: string
			IP address of the server. 'localhost' is also a valid address.
		port: integer
			Port of the server to which the requests have to be send.
		logging_path: string or `None`
			Location to where server output has to be logged. Will be additional to any other logging path defined via the logger module. If set to `None` no new logging path will be created and used.
		to_string_function: function or `None`
			Function used to convert PhenoAIResults instance to a string. See explanation above for more information. Can be set to `None` to use the default function. """

		global __serverinstance__
		logger.info("Starting server...")
		logger.info("Checking for valid PhenoAI instance")
		if port < 1025:
			raise exceptions.ServerException("Port of the server should be at least 1025.")
		server_address = (address, port)

		handler = PhenoAIServerRequestHandler
		if not utils.is_none(to_string_function):
			if not callable(to_string_function):
				raise exceptions.ServerException("Function provided to convert results object to string should be callable")
			else:
				class AlteredPhenoAIServerRequestHandler(PhenoAIServerRequestHandler):
					def convert_result_object_to_string(self, PhenoAIResults):
						return to_string_function(PhenoAIResults)
				handler = AlteredPhenoAIServerRequestHandler
		server = ThreadedHTTPServer(server_address, handler)


		__serverinstance__ = self
		if logging_path != None:
			if not utils.is_none(logger.__filechannel__):
				logger.warning("File logging will be continued in server log file '{}'".format(logging_path))
				logger.remove_file_channel()
			logger.info("Start logging to server log file")
			logger.to_file(logging_path)
		logger.warning("Server is running! Use <Ctrl-C> to stop")
		server.serve_forever()

	def run(self, data, map_data=False, ainalysis_ids=None, data_ids=None):
		""" Queries each added AInalysis for prediction on provided data

		This run method forms the core functionality of PhenoAI objects. It calls the run method of each of the added AInalyses with the provided data and configuration arguments. Returns are returned in a containers.PhenoAIResults object.

		The map_data argument is to be interpreted slightly differently here then the map_data argument of AInalysis objects. This argument controls if, for each of the AInalyses, the data is mapped before being subjected to the prediction methods of the estimator in the AInalysis. As such, this argument can either be True or False, just as the map_data argument of the AInalysis run method. However, this run method differs in the sense that is allows the entry "both", which queries all AInalyses that allow mapping of data *twice*: once with mapping and once without. These AInalyses thus have two AInalysisResults instances in the returned PhenoAIResults object, which can be differentiated based on their ID: the mapped results instance has "_mapped" appended to the ID name.

		Parameters
		----------
		data: numpy.ndarray, string or list of strings
			Data that has to be subjected to the estimator. Can be the raw data (numpy.ndarray), the location of a file to be read by the built-in file reader in the AInalysis or a list of file locations.
		map_data: boolean (default=False)
			Determines if data has to be mapped before prediction. Mapping uses the map_data method of this AInalysis and follows the mapping procedure defined in the configuration file for this AInalysis. Can also be "both", making each mapping-allowing AInalyses to be queried twice: once with and once without mapping. Mapped data returns an AInalysisResults object with "_mapped" appended to the AInalysisResults ID.
		ainalysis_ids: list of strings (default=`None`)
			If set to a list of strings, only AInalyses with an ID present in this list are queried for prediction on provided data. If set to `None`, all AInalyses are queried.
		data_ids: list or numpy.ndarray (default=`None`)
			List or numpy.ndarray of IDs for the data points. By setting this value, results can be extracted from the AInalysisResults object by using the IDs in this list/array instead of by its location in the result array. If `None`, this functionality will not be available. """

		# Use PhenoAI object locally
		# Create mapping iteration list
		mapmodes = []
		if map_data == "both":
			mapmodes.append(True)
			mapmodes.append(False)
			logger.info("Running PhenoAI with mapmode 'both'")
		else:
			mapmodes.append(bool(map_data))
			logger.info("Running PhenoAI with mapmode {}".format(bool(map_data)))

		# Create results object
		results = containers.PhenoAIResults()
		# Loop over ainalyses to request prediction
		for ainalysis in self.ainalyses:
			# Load estimator if not loaded already
			loaded = True
			if self.dynamic and not ainalysis.estimator.is_loaded():
				logger.debug("Loading estimator of AInalysis dynamically")
				ainalysis.estimator.load()
				loaded = False

			# Iterate over mapmodes
			for mapmode in mapmodes:
				# Test if this AInalysis should be queried
				if ainalysis_ids != None and ainalysis.ainalysis_id not in ainalysis_ids:
					continue

				# If multi mapmode skip mapmode True if AInalysis does not allow mapping
				if isinstance(ainalysis.configuration['mapping'], bool) and ainalysis.configuration['mapping'] == 0.0 and mapmode and len(mapmodes) > 1:
					continue
				logger.info("Running AInalysis '{}' in map mode '{}'".format(ainalysis.ainalysis_id, mapmode))

				# Do prediction
				logger.set_indent("+")
				result = ainalysis.run(data, map_data=mapmode, data_ids=data_ids)
				logger.set_indent("-")
				# Alter id if multi map mode
				if len(mapmodes) > 1:
					if mapmode:
						result.result_id += "_mapped"
				# Add result to AIResultContainer container
				results.add( result )

			# Unload estimator
			if not loaded:
				logger.debug("Clearing estimator of last AInalysis from memory")
				ainalysis.estimator.clear()

		logger.info("PhenoAI run finished, returning result")
		# Return results object
		return results


class ThreadedHTTPServer(ThreadingMixIn, HTTPServer):
	""" ThreadedHTTPServer implements multithreading for HTTP servers and is used to mutlithread HTTP requests for PhenoAI when run in server mode.

	Users do not have to interact with this method directly, it is created automatically and correctly when calling the run_as_server method of a PhenoAI instance. """
	pass

class PhenoAIServerRequestHandler(BaseHTTPRequestHandler):
	""" Handles HTTP requests for PhenoAI instances

	Users do not have to interact with this method directly, it is created automatically and correctly when calling the run_as_server method of a PhenoAI instance. """

	def do_GET(self):
		""" Takes care of the handling of HTTP GET requests made to PhenoAI

		Prediction requests to PhenoAI are to be made by POST request, this method returns a static text that is used to test the connection between client and server. Users dont have to interact with this method directly, it is automatically called when needed. """
		logger.info("Received GET request from {} - Connection availability is probably checked".format(self.client_address[0]))
		# Send response status code
		self.send_response(200)
		# Send headers
		self.send_header('Content-type','text/html')
		self.end_headers()
		# Write content as utf-8 data
		self.wfile.write(bytes("phenoai-ok :: Predictions can only be made via POST request.<br />Use the phenoai.client module or the C++ interface to do this easily.", "utf8"))
		return
	def do_POST(self):
		""" Takes care of the handling of HTTP POST requests made to PhenoAI

		Performs a prediction query to PhenoAI via its run method. Returns the resulting PhenoAIResults object in the correct format. Users dont have to interact with this method directly, it is automatically called when needed. """
		logger.info("Received POST request from {}".format(self.client_address[0]))
		logger.set_indent("+")
		try:
			# Get POST data
			content_length = int(self.headers['Content-Length'])
			post = self.rfile.read(content_length)
			post = post.decode('utf-8')
			post = cgi.parse_qs(post, keep_blank_values=1)
			for k, p in post.items():
				post[k] = p[0]
			# Split by mode
			if post["mode"] == "values":
				logger.debug("Received raw values")
				# Predict lists of values
				data = ast.literal_eval(post['data'])
				if post["data_ids"] == "false" or post["data_ids"] == "False":
					data_ids = None
				else:
					data_ids = ast.literal_eval(post["data_ids"])
				ainalysis_ids = ast.literal_eval(post["ainalysis_ids"])
				# Perform prediction
				if ainalysis_ids == "all":
					ainalysis_ids = None
				logger.debug("Calling run procedure of PhenoAI server instance")
				results = __serverinstance__.run(np.array(data), map_data=bool(float(post['mapping'])), ainalysis_ids=ainalysis_ids, data_ids=data_ids)
			elif post["mode"] == "file":
				logger.debug("Received file to be interpreted")
				data_ids = ast.literal_eval(post["data_ids"])
				ainalysis_ids = ast.literal_eval(post["ainalysis_ids"])

				# Create files in tmp folder
				filepath = "/tmp/{}.phenoai".format(utils.random_string(16))
				filepath_interpreted = "/tmp/{}_interpreted.phenoai".format(utils.random_string(16))

				logger.debug("Calling run procedure of PhenoAI server instance")
				try:
					with open(filepath, "w") as tmpfile:
						tmpfile.write(post['data'])
					results = __serverinstance__.run(filepath, map_data=bool(float(post['mapping'])), ainalysis_ids=ainalysis_ids, data_ids=data_ids)
					os.remove(filepath)
				except:

					with open(filepath_interpreted, "w") as tmpfile:
						#tmpfile.write(ast.literal_eval(post['data']))
						tmpfile.write(post['data'])
					results = __serverinstance__.run(filepath_interpreted, map_data=bool(float(post['mapping'])), ainalysis_ids=ainalysis_ids, data_ids=data_ids)
					os.remove(filepath_interpreted)

				if os.path.exists(filepath):
					os.remove(filepath)
				if os.path.exists(filepath_interpreted):
					os.remove(filepath_interpreted)
			else:
				raise exceptions.ServerException("Mode not recognized, should be either 'values' or 'file'. Provided was '{}'.".format(post['mode']))

			if "get_results_as_string" in post and float(post["get_results_as_string"]) == 1.0:
				logger.debug("Converting results object to string")
				# Return lists of results, not PhenoAIResults object
				results_txt = self.convert_result_object_to_string(results)
			else:
				# Convert to pickled instance
				logger.debug("Encoding results object to pickle")
				results_txt = io.pickle(results)
			returndict = {"status": "ok", "results": results_txt }
		except Exception as e:
			x = traceback.format_exc()
			logger.error(x)
			returndict = {"status": "error", "type":str(type(e).__name__), "message": str(e)}

		logger.info("Return results")
		# Send response status code
		self.send_response(200)
		# Send headers
		self.send_header('Content-type','text/html')
		self.end_headers()
		# Write content as utf-8 data
		returntext = json.dumps(returndict)
		self.wfile.write(bytes(returntext, "utf8"))
		logger.set_indent("-")
		return
	def convert_result_object_to_string(self, PhenoAIResults):
		""" Default function to convert a PhenoAIResults object to a string to be returned

		When no to_string_function is provided in the run_as_server method of the PhenoAI object but the client still requires a string to be returned instead of a PhenoAIResults object, this function is the default function to convert the PhenoAIResults object to a returnable string. It returns the results in .csv format with the data_ids in the first row (if provided). All subsequent rows contain the predictions for each of the data points.

		Parameters
		----------
		PhenoAIResults: PhenoAIResults object
			PhenoAIResults object that has to be converted to a string

		Returns
		-------
		outstr: string
			String version fo PhenoAIResults object. See description above for explanation of its contents."""
		out = None
		for i in range(len(PhenoAIResults)):
			preds = PhenoAIResults[i].get_predictions()
			if utils.is_none(out):
				if not utils.is_none(PhenoAIResults[i].data_ids):
					out = np.vstack((PhenoAIResults[i].data_ids, preds))
				if len(preds.shape) == 1:
					out = preds.reshape(1,-1)
				else:
					out = preds
			else:
				out = np.vstack((out, preds))
		out = [",".join(item) for item in out.astype(str)]
		outstr = "\n".join(out)
		return outstr



""" Functions to check for updates of PhenoAI and AInalyses

This module implements functions to check for updates of the PhenoAI package and individual AInalyses (if downloaded from the official PhenoAI website). This is done by sending an HTTP request to the PhenoAI server API. Returned information is automatically formatted in such a way that it is usable in Python.

All functions in this module are already implemented in various parts of the PhenoAI package. In normal circumstances the user should not need to call any of the functions in this module him- or herself.

!IMPORTANT! None of the functions in this module install anything automatically. This module is merely an update checker. Any updating has to be initiated manually by the user."""

from . import splash
splash.splash(False)

import sys
import importlib

from math import floor
import requests
import json

from . import __version__, __apiurl__
from . import exceptions
from . import logger
from . import utils



def check_ainalysis_update(uniquedbid, version):
	""" Check if there is a new version available for an AInalysis

	Checks if a new version is available on the PhenoAI server for the AInalysis with provided unique database ID. Versions are automatically checked for compatibility with versions of currently installed libraries (including PhenoAI) and the running python version. Querying of server is done by query_server function in this module.

	!IMPORTANT! This function does not automatically install the AInalysis.

	Parameters
	----------
	uniquedbid: string
		Unique ID of the AInalysis. Is stored in the configuration file of the AInalysis.
	version: integer
		Integer indicating the version number of the AInalysis.

	Returns
	-------
	error: boolean
		Boolean indicating if an error occurred on the server during processing of request. If True, return variable `text` will be the unformatted error message.
	update: boolean
		Boolean indicating if an update is available. If return variable `error` is True, this variable will be `None`.
	text: string
		Return message. If return variable `update` is True, this message will be formatted.
	disclaimer: string
		Disclaimer message on usage of premade AInalyses. """

	logger.debug("Querying server for ainalysis update information (dbid: {}, version: {})".format(uniquedbid, version))
	logger.set_indent("+")
	answer = query_server("ainalysis", {"ainalysis":uniquedbid, "current":version})
	logger.set_indent("-")
	return (answer["error"], answer["update"], format_server_message(answer["text"]))

def check_phenoai_update():
	""" Check if there is a new version available for the PhenoAI package

	Checks if a new version is available for the PhenoAI package. Versions are automatically checked for compatibility with versions of currently installed libraries and the running python version. Querying of server is done by query_server function in this module.

	!IMPORTANT! This function does not automatically install the new PhenoAI version.

	Returns
	-------
	error: boolean
		Boolean indicating if an error occurred on the server during processing of request. If True, return variable `text` will be the unformatted error message.
	update: boolean
		Boolean indicating if an update is available. If return variable `error` is True, this variable will be `None`.
	text: string
		Return message. If return variable `update` is True, this message will be formatted. """
	answer = query_server("phenoai")
	logger.set_indent("+")
	logger.debug("Querying server for phenoai update information (version: {})".format(__version__))
	logger.set_indent("-")
	return (answer["error"], answer["update"], format_server_message(answer["text"]))

def query_server(target, extra_post_data=None):
	""" Query server for information

	This function is called by the check_*_update functions in this module and should not be called directly by the user.

	Parameters
	----------
	target: string
		Select which update script to call on the server. Can be 'ainalysis' or 'phenoai'.
	extra_post_data: dictionary
		Add extra information to the dictionary that is sent to the server via POST request.

	Returns
	-------
	json_decoded: dictionary
		Dictionary of json decoded returned information. """
	postdata = {
		"phenoai": __version__,
		"python": "{}.{}.{}".format(sys.version_info[0], sys.version_info[1], sys.version_info[2])
	}
	for p in ["sklearn","tensorflow","keras"]:
		try:
			package = importlib.import_module(p)
			postdata[lib] = package.__version__
		except:
			pass

	if not utils.is_none(extra_post_data):
		postdata.update( extra_post_data )

	try:
		data = requests.post(__apiurl__.format(target), data=postdata, timeout=5)
		logger.debug("Data returned from server: {}".format(data.text))
		j = data.json()
		return j
	except requests.exceptions.ConnectionError:
		logger.debug("Could not create connection with PhenoAI update server")
		return (False, None, "Could not create a connection to the PhenoAI server.")
	except json.decoder.JSONDecodeError:
		logger.debug("Server response was incorrectly formatted")
		return (False, None, "Got invalid formatted response from the server.")


def format_server_message(text, border=True, line_length=75):
	""" Format server message to be centered with a certain line length

	Formats the provided text to be centered in lines of provided length. If the text is longer than the provided length, the text will continue on new lines until the text stops.

	Parameters
	----------
	text: string
		Text that will be centered and put over multiple lines.
	border: boolean (default=`True`)
		If set to `True`, the first and last line of the new text will consist out of line_length * "="
	line_length: integer (default=75)
		Number of characters per line

	Returns
	-------
	formatted_text: string
		Formatted text. Line breaks are denoted by \n """
	lines = []
	text = text.strip()
	words = text.split()
	if border:
		lines.append("="*line_length)
	line = ""
	i = 0
	for word in words:
		if len(line)+len(word)+1 <= line_length:
			line += " {}".format(word)
		if len(line)+len(word)+1 > line_length or i == len(words)-1:
			spaces = floor((line_length - len(line))/2)
			for s in range(spaces):
				line = " {}".format(line)
			lines.append(line)
			line = ""
		i += 1
	if border:
		lines.append("="*line_length)
	return "\n".join(lines)
